#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
program: _alan_str.py
Description: ALAN string utility functions for jinja2
Version: 0.91
Functions:
  prerr(e,*args,**kargs):
  str_contains(va=[],subs=[],vb=[],uniqueTF=True):
  str_matches(va=[],subs=[],vb=[],uniqueTF=True):
  clean_punctuation2upper(s,c=' '):
  remove_tags(raw_html):
  unique_set(a):
  check_latest_macro(cdt=None,nwk=1):
  dLst2df(dx,keyName='ticker',valName='pctChange'):
  merge_t2l(tkLst=[],df={},colx='ticker',**optx):
  merge_yqc(tkLst=[],df={},colx='ticker',**optx):
  merge_stock_info(tkLst=[],df={},colx='ticker',how='left',**optx):
  jsonCvt(x):
  dtCvt(x):
  jsonCvt0(x):
  jsonObjReload(zobj,cvtFunc=jsonCvt0):
  hourly_eod(ticker,start=None,mp3YN=False,src='yh',lang='cn',end_hm=1600,dirname='templates/',outdir='US/mp3_hourly',**optx):
  syscall_eod(ticker,start=None,mp3YN=False,src='yh',wkDir='.'):
  find_numbers(s,convertTF=False,pattern='',prefix='',suffix=''):
  popen(args,s=None,shell=True,bufsize=1024,stdin=PIPE,stdout=PIPE,stderr=PIPE,debugTF=False,**optx):
  sysCall(xcmd=None):
  popenCall(xcmd=None,xin=None,bufsize=1024,shell=True,debugTF=False):
  str2float(s,r=None):
  strc2float(s,r=None,sc=','):
  str2value(s,r=None):
  num_en2cn(s='10B',numTF=False):
  sub_special_char(s='',p='{}[]()#%|*,:;/ ',n=''):
  consecutive_sign(v,zeroSign=0,reverseTF=True):
  run_jj2(ts,dd={},dirname='.',fileTF=False,debugTF=False):
  jj2_fmt(ts,dirname='.',fileTF=False,debugTF=False,j2onlyTF=False,**dd):
  get_arg2func(funcName,**optx):
  get_funcLst(unLst=['x_','Call','OLD','_DEPRECATED','Popen']):
  jj_fmt(ts,dd={},dirname='.',fileTF=False,debugTF=False,j2onlyTF=False,**opts):
  ifelse(yn,a,b):
  clip(a,x,b):
  isclip(a,x,b):
  isclipL(a,x,b):
  isclipR(a,x,b):
  num2MP3(s,dotSign='點',lang="cn",mp3YN=True):
  gsrg(r,n):
  roundTBM(r,n):
  roundSD(r,n,lang='cn',mp3YN=False,TBMTF=False): 
  roundPctMP3(r,n=0,pct=100,dotSign='.',lang='en'):
  roundPct(r,n=0,pct=100,dotSign='.',lang='en',mp3YN=False):
  loc_dindex(r,thd=[1.96,1.15,0.32,0.1,-0.1,-0.32,-1.15,-1.96],openTF=True,sortTF=False):
  loc_aindex(r,thd=[-1.96,-1.15,-0.32,-0.1,0.1,0.32,1.15,1.96],openTF=True,sortTF=False):
  udfSentence(tkName='AAPL',tkVal=999,tkChg=0,tkChgPct=0,wordLst=None,thd=None,lang='cn',openTF=True,sortTF=False,ascendingTF=False,scale=1,pct=1,mp3YN=False):
  udfWord(r,wordLst=None,thd=None,lang='cn',openTF=True,sortTF=False,ascendingTF=False,scale=1):
  performance_verb(r,udfLst):
  btwDates(a,b,bracket=None,fmt=None,lang="en"):
  btwDates_cn(a,b,bracket=None,fmt=None):
  btwDates_en(a,b,bracket=None,fmt=None):
  datetime2hm(dtm,dformat="%H:%M",lang="cn",mp3YN=False):
  ymd2datetime(s,dformat="%Y%m%d"):
  ymd2ym(s,ym=None,ymd='%Y-%m-%d',lang="cn"):
  ymd2ymd(s,ym=None,ymd='%Y-%m-%d',lang="cn",):
  ymd2md(s,ym='%m-%d',ymd='%Y%m%d',lang="en",mp3YN=False):
  pattern_cn(name,lang="cn"):
  fq2unit_str(fqStr,lang="cn"):
  rdWord(wordLst,sep="|"):
  udfSet(udf=None,lang="en",tense="past"):
  udfSym(r,udf=[u'\u25B2',u'\u25BC',''],zs=0):
  udfStr(r,udf=None,zs=[0,0],lang="en",tense="past"):
  udfStr_DEPRECATED(r,udf=None,zs=0,lang="en",tense="past"):
  roundMP3(x,n=0,usdSign=None,dotSign='.',lang=None):
  roundAuto(x,n=2,usdSign=None,dotSign='.',lang='en',mp3YN=False):
  roundUSD(x,n=0,usdSign='$',dotSign='.',lang='en',mp3YN=False):
  wrapWord(a,bk=["(",")"]):
  rangeStr(a,b,bk=["[","]"]):
  btwStr(a,b,bk=["ranging between","and"],fl="fluctuated around"):
  btwStr_cn(a,b,bk=["介于","至","的範圍之間"],fl=["變化不大，只會在","內上下稍稍波動"],unitStr=''):
  connMdb(clientM=None):
  find_mdb(jobj={},clientM=None,dbname='ara',tablename=None,field={},sortLst=[],ascendingTF=False,limit=0,dfTF=False,**optx):
  get2mndb(jobj,tablename='lsi2nlg',zpk={'ticker','username','tmplname','category','tmpltype','lang'}):
  upsert_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,wmode='upsert',zpk=[],zasc=[],**optx):
  write_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,wmode='upsert',zpk=[],zasc=[],**optx):
  delete_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,**optx):
  insert_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,wmode='',zpk=[],zasc=[],**optx):
  save2mgdb(mLst=None,mky=None,dbM=None,tablename="temp",wmode='replace'):
  upd2many(dd=[],zpk={},mcur={},debugTF=False):
  upd2mdb(dd=[],zpk={},mcur={},debugTF=False):
  np2mongo(zobj):
  write2mdb(jobj,clientM=None,dbname='ara',tablename=None,zpk={},insertOnly=False):
  df_tofile(fp,df,output=None,sep='|',indexTF=False,hdrTF=True):
  str_tofile(fp,s):
  df_output(df,output=None,sep='|',indexTF=False,hdrTF=True):
  save2mndb(jobj,zpk={'ticker','username','tmplname','category','tmpltype','lang'},clientM=None,dbname='ara'):
  lst2dict(jkst,jlst):
  tmpl2lsi(tmplstr,argobj):
  combine_cmd2dict(*_xargs_): # TBD ,**_xdu_):
  jobj2lsi(jobj):
  lsi2nlg_calc(jobj):
  qs_split(xstr,d1='&',d2='='):
  qs_exec(xstr):
  qs_execOLD(xstr):
  extra_opts(opts={},xkey='',method='JS',updTF=True):
  generate_comment_hourly(ts,df,opts=None):
  trend_forecast(ticker='^GSPC',freq='W',debugTF=False,pgDB=None,**optx):
  stock_screener(scrIds='most_actives',sortLst=['dollarValue'],ascLst=[False],**optx):
  x_screener(nobs=6,sortLst=['dollarValue'],ascLst=[False],dbname='ara',tablename='yh_qutoe_curr',addiFilter=1,debugTF=False,colx=[],addiClause='',**optx):
  ssh2mg(sshUser='rstudio',rhost='api1.beyondbond.com',lhostLst=['bbapi1','api1'],port=27017,**optx):
  cgi2dict(**optx):
Last Mod., 
Wed May 27 13:45:46 EDT 2020
Tue Aug  3 16:25:32 EDT 2021
Wed Aug 10 12:41:28 EDT 2022
+ add ssh2mg()
Thu Aug 18 14:23:48 EDT 2022
+ rename upd2mdb() to upd2mdbOLD()
+ add upd2mdb()
+ add upd2many()
Fri Sep  2 17:35:32 EDT 2022
+ add cgi2dict()
Fri Sep  9 10:25:12 EDT 2022
"""

#from __future__ import print_function
import sys
import os
from jinja2 import Environment as jinEnv, FileSystemLoader
import jinja2.ext
import pickle
from pymongo import MongoClient
from bson.binary import Binary
from string import Formatter
from math import *
import random
import locale
import datetime
import types
import json
import numpy as np
import pandas as pd
from pandas import Timestamp
import re
from subprocess import Popen, PIPE, STDOUT
from importlib import import_module
# IMPORT external functions
from _alan_date import next_date,ymd2dt,dt2ymd,s2dt
from _alan_calc import saferun,subDict,subDF,subVDict,renameDict,getKeyVal,sqlQuery,conn2mgdb,safeRunArg

if sys.version_info.major == 2:
	reload(sys)
	sys.setdefaultencoding('utf8')

def prerr(e,*args,**kargs):
	'''print error message
	e.g.,
	prerr("wrong variable")
	'''
	import sys
	import io
	file=kargs.pop('file',sys.stderr)
	if len(args)>0:
		s=kargs.pop('s','**ERROR: {0} @{1}()')
	else:
		s=kargs.pop('s','**ERROR: {0}')
	ret=s.format(e,*args,**kargs)
	if isinstance(file,io.IOBase):
		print(ret,file=file)
	return ret

def str_contains(va=[],subs=[],vb=[],uniqueTF=True):
	'''
	search a string-list:va that contains any of string-list:subs,
	then add addtional string-list vb
	then return a unique (optional) list of strings
	'''
	v = [x for x in va if any(map(x.__contains__,subs))]+vb
	return [set(v)] if uniqueTF else v

def str_matches(va=[],subs=[],vb=[],uniqueTF=True):
	'''
	search a string-list:va that matches any of string-list:subs,
	then add addtional string-list vb
	then return a unique (optional) list of strings
	'''
	v = [x for x in va if x in subs]+vb
	return [set(v)] if uniqueTF else v

def clean_punctuation2upper(s,c=' '):
	from string import punctuation
	xs = s.translate(s.maketrans(punctuation,c*32))
	xs = re.sub('\s+',' ',xs).strip().upper()
	return xs

def remove_tags(raw_html):
	''' remove HTML tags
	'''
	cleanr = re.compile('<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});')
	cleantext = cleanr.sub('', raw_html)
	#cleantext = re.sub(cleanr, '', raw_html)
	return cleantext

def unique_set(a):
	''' return unique set with the same order
	'''
	return sorted(set(a), key=a.index)

def check_latest_macro(cdt=None,nwk=1):
	'''
	return macro info in dict via checking 
	if latest macro event is within last 2 weeks since last Monday
	otherwise return None
	Note, 
	  If annualRate is calc when 'deriv' is added in 'mapping_series_label' source column 
	  'annualRate' will be added and pctChange will apply to 'annualRate' variable
	  rather than 'value'
	'''
	from _alan_date import tg_latest2week,ymd2dt
	xqr="""SELECT  x.value,x.pbdate,y.* from macro_hist_fred x, (SELECT a.vntdate,b.* FROM macro_vintage_date a, mapping_series_label b
where a.series=b.series and b.freq in ('M','Q') ORDER BY vntdate DESC limit 1) as y
WHERE x.series=y.series ORDER BY pbdate DESC limit 13 """
	#df = sqlQuery(open('latest_macro_vintage.sql').read())
	df = sqlQuery(xqr)
	in2wTF = tg_latest2week(ymd2dt(df['vntdate'].iloc[0]),cdt=cdt,nwk=nwk)

	#=  Check If annualRate is calc when source='deriv' in the 'mapping_series_label' table
	dg=sqlQuery("SELECT * from macro_hist_fred where series='{}' ORDER BY pbdate".format(df['series'].iloc[-1]+"_PCTCHG"))
	if len(dg)>0:
		dg.rename(columns={"value":"annualRate"},inplace=True)
		df = df.merge(dg[['annualRate','pbdate']],on='pbdate')
		df['pctChange'] = df['annualRate'].diff(-1)
	else:
		df['pctChange'] = df['value'].pct_change(-1)

	if in2wTF is not None and len(in2wTF)>0:
		in2wTF.update(df.iloc[0].to_dict())
		in2wTF.update(f=df)
	return in2wTF


def dLst2df(dx,keyName='ticker',valName='pctChange'):
	'''
	convert list of dictionay to dataframe
	'''
	if len(dx)<1:
		return {}
	df=pd.DataFrame([list(x.items())[0] for x in dx],columns=[keyName,valName]).set_index(keyName,drop=True)
	return df

def merge_t2l(tkLst=[],df={},colx='ticker',**optx):
	'''
	Merge ticker2label info into dataFrame 'df' based on 'colx' column
	where 'colx' default to 'ticker'
	'''
	from ticker2label import ticker2label
	kwargs=dict(output=None,funcArg=ticker2label)
	kwargs.update(optx)
	dm = merge_stock_info(tkLst,df,**kwargs)
	return dm

def merge_yqc(tkLst=[],df={},colx='ticker',**optx):
	'''
	Merge stock quote info into dataFrame 'df' based on 'colx' column
	where 'colx' default to 'ticker'
	'''
	from yh_chart import yh_quote_comparison
	kwargs=dict(screenerTF=False,funcArg=yh_quote_comparison)
	kwargs.update(optx)
	dm = merge_stock_info(tkLst,df,**kwargs)
	return dm

def merge_stock_info(tkLst=[],df={},colx='ticker',how='left',**optx):
	from ticker2label import ticker2label
	funcArg=getKeyVal(optx,'funcArg',ticker2label)
	kwargs={}
	kwargs.update(optx)
	kwargs.pop('funcArg',None)
	dq = funcArg(tkLst,**kwargs)
	if len(dq)<1 or colx not in dq:
		return df
	if colx not in df:
		return dq
	return df.merge(dq,on=colx,how=how)

def jsonCvt(x):
	return dtCvt(x)

def dtCvt(x):
	from bson.objectid import ObjectId
	import pandas as pd
	import datetime
	if isinstance(x, (ObjectId,datetime.datetime,pd.Timestamp)):
		return x.__str__()
	elif isinstance(x, (pd.DataFrame)):
		return x.to_dict(orient='records')
	elif isinstance(x, np.ndarray):
		return list(x)
	elif isinstance(x, np.integer):
		return int(x)
	elif isinstance(x, np.float):
		return float(x)
	return x

def jsonCvt0(x):
	import pandas as pd
	if isinstance(x, (pd.DataFrame)):
		return x.to_dict(orient='records')
	elif isinstance(x, np.ndarray):
		return list(x)
	elif isinstance(x, np.integer):
		return int(x)
	elif isinstance(x, np.float):
		return float(x)
	return x

def jsonObjReload(zobj,cvtFunc=jsonCvt0):
	from bson.json_util import dumps, loads
	if cvtFunc is None:
		return zobj
	try:
		zobj=loads(dumps(zobj,default=cvtFunc))
	except Exception as e:
		sys.stderr.write("**ERROR:{} @{}\n".format(str(e),"jsonObjReload"))
	return zobj

def hourly_eod(ticker,start=None,mp3YN=False,src='yh',lang='cn',end_hm=1600,dirname='templates/',outdir='US/mp3_hourly',**optx):
	'''
	hourly EoD run
	'''
	from ticker2label import ticker2label
	from hourly_mkt import wrap_hourly 
	tkInfo = ticker2label(ticker)
	if len(tkInfo)>0:
		optx.update(tkInfo)
	cname = 'company_cn' if lang == 'cn' else 'company'
	title=optx[cname] if cname in tkInfo  else ticker
	optx.update(start=start,mp3YN=mp3YN,src=src,lang=lang,dirname=dirname,title=title,outdir=outdir)
	end_hm= clip(900,end_hm,1600)
	sys.stderr.write("=====Running {}:{} @{}\n".format(start,end_hm,"hourly_eod"))
	jobj = wrap_hourly([ticker],end_hm=end_hm,**optx)
	if 'hourlyObj' in optx and isinstance(optx['hourlyObj'],dict):
		# Static setup fields
		fields=['ticker','comment','mp3comment','change','changePercent','copen','close','cprice','xprice','cpchg','marketCap','sector','title','rpt_hm','rpt_time','f']
		# Dynamic setup fields
		#fields = [ky for ky,val in jobj.items() if isinstance(val,(str,int,float)) ]
		#fields.append('f')
		fieldTyp = {ky:type(val) for ky,val in jobj.items() if ky in fields}
		sys.stderr.write(" --field/type saved for hourlyObj:{}\n".format(fieldTyp))
		optx['hourlyObj'].update(subDict(jobj,fields))
	return jobj['mp3comment'] if mp3YN else jobj['comment']

# DEPREATED, use hourly_eod
def syscall_eod(ticker,start=None,mp3YN=False,src='yh',wkDir='.'):
	'''
	shortcut systemCall for EoD run
	'''
	from ticker2label import ticker2label
	hourlyTmp='(cd {wkDir}; hourly_mkt.py  --start={start} --extra_xs="archiveTest=True;mp3YN={mp3YN};mp3TF=False;target_hm=[1600];mp3TF=False;plotTF=False;outTF=True" {ticker} --src={src} --title="{company_cn}" )'
	tkInfo = ticker2label(ticker)
	if len(tkInfo)<1:
		tkInfo=dict(company_cn=ticker,ticker=ticker)
	xstr = hourlyTmp.format(wkDir=wkDir,start=start,mp3YN=mp3YN,src=src,**tkInfo)
	o,e = popenCall(xstr)
	sys.stderr.write("{} {}\n".format(o,e))
	return o

def find_numbers(s,convertTF=False,pattern='',prefix='',suffix=''):
	'''
	extract all the numbers contained in a string
	and convert them into float if 'convertTF' is True
	'''
	if not pattern:
		pattern="[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?"
	if prefix:
		pattern=prefix+pattern
	if suffix:
		pattern=pattern+suffix
	rr = re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", s)
	for j,x in enumerate(rr):
		if ',' in x :
			rr[j] = x.replace(',','')
		elif prefix and prefix in x:
			rr[j] = x.replace(prefix,'')
		elif suffix and suffix in x:
			rr[j] = x.replace(suffix,'')
	if convertTF:
		rr = [ float(x) for x in rr]
	return rr

def popen(args,s=None,shell=True,bufsize=1024,stdin=PIPE,stdout=PIPE,stderr=PIPE,debugTF=False,**optx):
	'''
	System call via subprocess.Popen
	Where
	  args: can be a string of shell command is shell=True or
		a list of command args to form shell comman
	  s:	is the input string for stdin if there is any
	'''
	if debugTF:
		sys.stderr.write("==SHELL_CMD:\n{}\n".format(args))
	p = Popen(args,shell=shell,bufsize=bufsize,stdin=stdin,stdout=stdout,stderr=stderr)
	xin = (s+"\n").encode('utf-8') if s else None
	o,e = p.communicate(input=xin)
	if debugTF==2:
		msgerr="===SHELL_CMD:\n{}\n---OUT:\n{}\n---ERR:\n{}\n".format(args,o.decode(),e.decode())
		sys.stderr.write(msgerr)
	return o,e

def sysCall(xcmd=None):
	""" run [xcmd] as system call  """
	return Popen(xcmd, shell=True, stdout=PIPE).stdout.read()

def popenCall(xcmd=None,xin=None,bufsize=1024,shell=True,debugTF=False):
	""" Description: system call via subprocess.Popen """
	if not xcmd:
		return None,None
	if debugTF:
		sys.stderr.write("==SHELL_CMD:\n{}\n".format(xcmd))
		if xin:
			sys.stderr.write("  input:\n{}\n".format(xin))
	p = Popen(xcmd,shell=shell,bufsize=bufsize,stdin=PIPE,stdout=PIPE,stderr=PIPE)
	xin = (xin+"\n").encode('utf-8') if xin else None
	o, e = p.communicate(input=xin)
	return o.decode(), e.decode()

def str2float(s,r=None):
	'''
	convert string 's' to a valid float or 'r'=None
	'''
	try:
		return float(s)
	except Exception as e:
		return r

def strc2float(s,r=None,sc=','):
	'''
	convert comma string 's' to a valid float or 'r'=None
	
	'''
	if len(sc)>0:
		s = re.sub('[{}]'.format(sc),'',s)
	return str2float(s,r)

def str2value(s,r=None):
	'''
	convert string 's' to a valid numeric value or 'r'=None
	'''
	v = re.search("([+-]?\d*[.]?\d*)([eE][+-]?\d\d*)?",s).groups()
	if len(v[0])>0:
		return str2float(s,r)
	else:
		return r

def num_en2cn(s='10B',numTF=False):
	''' convert 's' from '10B' into '100億'
	'''
	v = re.search("[+-]?(\d*[.]?\d*)\s?(\w)",s)
	n,u = v.groups()
	u = u.upper()
	n = float(n)
	if u == 'B':
		x = "{:.0f}{}".format(n*10,'億')
		n = n*10**9
	elif u == 'M':
		x  = "{:.0f}{}".format(n*100,'萬')
		n = n*10**6
	elif u == 'T':
		x = "{}{}".format(n,'兆')
		n = n*10**12
	else:
		x = "{}".format(n)
	if numTF:
		return n
	return x
	

def sub_special_char(s='',p='{}[]()#%|*,:;/ ',n=''):
	"""
	covert string 's' that contains character in 'p' string and replace each one with 'n'
	"""
	import re
	rx = '[' + re.escape(''.join(p)) + ']'
	return re.sub(rx,n, s)

def consecutive_sign(v,zeroSign=0,reverseTF=True):
	"""
	Calc (number of same consecutive sign, sign) from the end
	Note,
	zeroSign can be treated as [-1,0,1]
	"""
	if len(v)<2:
		return 1,0
	if reverseTF is True:
		xv = v[::-1]
	else:
		xv = v
	if zeroSign==1:
		xv=np.where(np.array(xv)>=0,1,-1)
	elif zeroSign==-1:
		xv=np.where(np.array(xv)>0,1,-1)

	s = np.sign(xv).astype(int)
	b = s[0]
	for j,x in enumerate(s[1:]):
		if x!=b:
			j=j+1
			break
	return j, b

def run_jj2(ts,dd={},dirname='.',fileTF=False,debugTF=False):
	"""
	run jinja2 template string [ts] to output text [jj_str]
	where: 
		ts: template string
		**dd:  rendered inputs in {}
	Note: 
		[ts] is a template filename in [dirname] if fileTF is True
	"""
	#from jinja2 import Template as jjTemplate
	j2Env = jinEnv(loader=FileSystemLoader(dirname),trim_blocks=True,extensions=['jinja2.ext.do'])
	if debugTF:
		sys.stderr.write("=====run_jj2():\nfileTF:{},debugTF:{}\nts:{}\n".format(fileTF,debugTF,ts) )
		sys.stderr.write("dd:{}\n".format(dd) )
	if not ts:
		''
	if fileTF is False:
		return(j2Env.from_string(ts).render(**dd))
		#return jjTemplate(ts).render(**dd)
		#no loader environment specified for jjTemplate
	else:
		return(j2Env.get_template(ts).render(**dd))

def jj2_fmt(ts,dirname='.',fileTF=False,debugTF=False,j2onlyTF=False,**dd):
	return jj_fmt(ts,dd,dirname=dirname,fileTF=fileTF,debugTF=debugTF,j2onlyTF=j2onlyTF)

def get_arg2func(funcName,**optx):


	'''
	return a function-call or None based on 'funcArg'
	'''
	funcArg=None
	if isinstance(funcName,str):
		if funcName in globals() and hasattr(globals()[funcName],'__call__'):
			funcArg =  globals()[funcName]
	return funcArg

def get_funcLst(unLst=['x_','Call','OLD','_DEPRECATED','Popen']):
	henv={ k:v for (k,v) in globals().items() if callable(v) and not any(map(k.__contains__,unLst))}
	return henv

@saferun
def jj_fmt(ts,dd={},dirname='.',fileTF=False,debugTF=False,j2onlyTF=False,**opts):
	"""
	run jinja2 template string [ts] to return output text [jj_str]
	based on input **dd + buildins + _alan_str defined functions
	where: 
		ts: template string
		dd:  rendered inputs in {}
	Note: 
		[ts] is a template filename in [dirname] if fileTF is True
	"""
	try:
		if ts is None:
			return ''
		elif len(ts)<1:
			return ''
		inisd = int(datetime.datetime.today().strftime("%m%d")) + hash(ts)%10000
		random.seed(inisd)
		#- Add available embedded function
		#unLst = ['sysCall','popenCall','_DEPRECATED','x_']
		#tpLst = (types.BuiltinFunctionType,types.FunctionType)
		#genv = { k:v for (k,v) in globals().items() if isinstance(v,tpLst) is True }
		#henv = {ky:va for ky,va in genv.items() if not any(map(ky.__contains__,unLst))}
		henv = get_funcLst()

		# Inherit variable from 'dd':
		d = dd.copy()
		# Add built-in functions:
		d.update(__builtins__)
		# Add Internal functions inside '_alan_str':
		d.update(henv) # include all the functions
		# Add additional variables from **opts:
		if len(opts)>0:
			d.update(opts)
		# Update debugTF and dirname variables:
		d.update(debugTF=debugTF,dirname=dirname)
		ctrlSTRUCT=getKeyVal(opts,'ctrlSTRUCT','')
		if ctrlSTRUCT is not None and ctrlSTRUCT in ['include','extends']:
			ts='{{% {} "{}" %}}'.format(ctrlSTRUCT,ts)
		if debugTF:
			sys.stderr.write( "@: {}\n".format(datetime.datetime.now()) )
			sys.stderr.write( "jj2 keys: {}\n".format(sorted(d.keys())) )
			sys.stderr.write( "Interal Func:\n{}\n".format(sorted(henv.keys())) )
			sys.stderr.write( "jj2 ts:\n{}\n".format(ts) )
			sys.stderr.write( "jj2 d:\n{}\n".format(d) )
		# Need at least one of tags ['{{' , '{%' ] in ts to run jinja2
		if '{{' in ts or '{%' in ts or fileTF is True or j2onlyTF is True:
			jj_str =  run_jj2(ts,dd=d,dirname=dirname,fileTF=fileTF,debugTF=debugTF)
		else: 
			jj_str = ts.format(**d)

		if jj_str is None:
			jj_str=''
		elif hasattr(jj_str,'__len__') and len(jj_str)>0:
			jj_str = ''.join([s for s in jj_str.splitlines(True) if s.strip()])
	except Exception as e:
		fnx=sys._getframe().f_code.co_name
		jj_str = "**ERROR:{} @{}()".format(e,fnx)
		jj_str += "\nTS:\n{}{}".format(ts[:40], " ..." if len(ts)>40 else "")
	return(jj_str)

def ifelse(yn,a,b):
	return a if yn else b

def clip(a,x,b):
	return min(max(a,x),b)

def isclip(a,x,b):
	return (True if a<=x and x<=b else False)

def isclipL(a,x,b):
	return (True if a<=x and x<b else False)

def isclipR(a,x,b):
	return (True if a<x and x<=b else False)

def num2MP3(s,dotSign='點',lang="cn",mp3YN=True):
	""" convert number str to audio readable format
	"""
	#TBD, buggy codes
	#if lang=='cn':
	#	dotSign='點'
	#	s = num_en2cn(s)
	s=str(s).replace(',','').strip()
	if '.' not in s or mp3YN==False:
		return s
	else:
		xs=s.split('%')
		pctSign='%' if len(xs)>1 else ''
		x=xs[0].split('.')
		if int(x[1])==0:
			return "{}{}".format(x[0],pctSign)
		if lang=="cn": 
			dotSign='點'
			y="{0}{3} {1} {2}".format(x[0],dotSign," ".join(x[1]),pctSign)
		else:
			dotSign="point"
			y="{0} {1} {2}{3}".format(x[0],dotSign," ".join(x[1]),pctSign)
		return y

def gsrg(r,n):
	""" get signicant_rounding digit """
	return int(max(ceil(log(abs(r),.1)),n)) if abs(r)>0 else n

def roundTBM(r,n):
	""" 
	rounding r to nearest [T|B|M] and its (n, signicant_digit)
	"""
	if r>10**12:
		xstr="{:.{}f}T".format(r/10.0**9,n)
	elif r>10**9:
		xstr="{:.{}f}B".format(r/10.0**9,n)
	elif r>10**6:
		xstr="{:.{}f}M".format(r/10.0**6,n)
	else:
		xstr="{:.{}f}".format(r,n)
	return xstr

def roundSD(r,n,lang='cn',mp3YN=False,TBMTF=False): 
	""" rounding r to max (n, signicant_digit)
	"""
	n=gsrg(r,n)
	if TBMTF:
		xstr = roundTBM(r,n)
	else:
		xstr="{:.{}f}".format(r,n)
	if mp3YN==True:
		xstr=num2MP3(xstr,lang=lang,mp3YN=mp3YN)
	return xstr

def roundPctMP3(r,n=0,pct=100,dotSign='.',lang='en'):
	n=gsrg(r*pct,n)
	s="{:.{}f}%".format(r*pct,n)
	if n<1:
		return s
	else:
		x=s[:-1].split(".")
		if int(x[1])==0:
			return "{}{}".format(x[0],"%")
		if lang=="cn":
			dotSign='點'
			y="{} {} {}{}".format(x[0],"%",dotSign," ".join(x[1]))
		else:
			y="{} {} {}{}".format(x[0],dotSign," ".join(x[1]),"%")
		return y

def roundPct(r,n=0,pct=100,dotSign='.',lang='en',mp3YN=False):
	"""
	return (r*100)% rounding to nearest n decimal
	Note,
		dotSign prints decimal point '.' for mp3 sound
		if mp3 flag mp3YN is True
	"""
	if mp3YN is True:
		return roundPctMP3(r,n=n,pct=pct,dotSign=dotSign,lang=lang)
	n=gsrg(r*pct,n)
	return "{:.{}f}%".format(r*pct,n)

def loc_dindex(r,thd=[1.96,1.15,0.32,0.1,-0.1,-0.32,-1.15,-1.96],openTF=True,sortTF=False):
	"""
	Locate the 1st index of descending thresholds:thd for r > thd[index] 
	Return:
		index location
	Note:
		return the 1st index of r >= thx[index] if openTF = False
	"""
	n = len(thd)
	if sortTF is True:
		thd = np.sort(thd)[::-1]
	if openTF is False:
		#ul = filter(lambda x: thd[x]-r <=0, range(n))[:1]
		ul = [ x for x in range(n) if thd[x]-r<=0 ][:1]
	else:
		#ul = filter(lambda x: thd[x]-r < 0, range(n))[:1]
		ul = [ x for x in range(n) if thd[x]-r<0 ][:1]
	return ul[0] if len(ul)>0 else n

def loc_aindex(r,thd=[-1.96,-1.15,-0.32,-0.1,0.1,0.32,1.15,1.96],openTF=True,sortTF=False):
	"""
	Locate the 1st index of ascending thresholds:thd for r > thd[index] 
	Return:
		index location
	Note:
		return the 1st index of r >= thx[index] if openTF = False
	"""
	n = len(thd)
	if sortTF is True:
		thd = np.sort(thd)
	if openTF is True:
		ul = list(filter(lambda x: thd[x]-r >=0, range(n)))[:1]
	else:
		ul = list(filter(lambda x: thd[x]-r > 0, range(n)))[:1]
	return ul[0] if len(ul)>0 else n

def udfSentence(tkName='AAPL',tkVal=999,tkChg=0,tkChgPct=0,wordLst=None,thd=None,lang='cn',openTF=True,sortTF=False,ascendingTF=False,scale=1,pct=1,mp3YN=False):
	"""
	Return directional descripting sentence based on threshold: thd 
	Inputs: tkName,tkVal,tkChg,tkChgPct(in %)
	Example: udfSentence(tkName='道瓊',tkVal=35000,tkChg=2,tkChgPct=-1.2)
	"""
	dx=locals().copy()
	tmplstr_cn="""
{{tkName}}{{udfWord(tkChgPct,lang=lang)}}
{%- if abs(tkChgPct)>0.5 -%}
	{{roundPct(tkChgPct,0,pct=pct,lang=lang,mp3YN=mp3YN)}}
{%- elif abs(tkChg)>=0.5 -%}
	{{roundSD(tkChg,0,lang=lang,mp3YN=mp3YN)}}點
{%- endif -%}
，{{rdWord('來到|達|為')}}{{roundSD(tkVal,0,lang=lang,mp3YN=mp3YN)}}。
"""
	tmplstr_en="""
{{tkName}} {{udfWord(tkChgPct,lang=lang)}} {{''}}
{%- if abs(tkChgPct)>0.5 -%}
	{{roundPct(tkChgPct,0,pct=pct,lang=lang,mp3YN=mp3YN)}}
{%- elif abs(tkChg)>=0.5 and abs(tkChgPct)>0.032 -%}
	{{roundSD(tkChg,0,lang=lang,mp3YN=mp3YN)}}
{%- endif -%}
, {{rdWord('reached to|at')}} {{roundSD(tkVal,0,lang=lang,mp3YN=mp3YN)}}. 
"""
	tmplstr=tmplstr_cn if lang=='cn' else tmplstr_en
	return jj_fmt(tmplstr,dx)

def udfWord(r,wordLst=None,thd=None,lang='cn',openTF=True,sortTF=False,ascendingTF=False,scale=1):
	""" 
	Directional descriptor of r based on threshold: thd 
	Where
	  r: float or int value
	  thd[]: threadhold numeric array
	  wordLst[]: a list of action word w.r.t thd[index]
	Return:
	  Directional verb/adjective
	Note:
	  wordLst[] array should be 1+ length of thd[]
	  thd: threshold buckets
	  default to [1.96,1.15,0.16,0.032,-0.032,-0.16,-1.15,-1.96] multiplied by 'scale'
	Example: udfWord(1.4,lang='cn')
	"""
	if thd is None:
		thd = [1.96,1.15,0.16,0.032,-0.032,-0.16,-1.15,-1.96]
		if scale != 1:
			thd = [x*scale for x in thd ]
	if wordLst is None and lang=='cn':
		xstr="""
強勢上揚大漲|大幅升高走強|強勢攀升上漲|大幅上揚升高
明顯走高|明確向上攀升|強勢上揚
上升|攀升|上漲
小幅上漲|微漲|微微攀升|小幅微升
上下微動|輕微震盪|區間微震，幾無變化|稍稍波動，幾無變化|沒什麼變動
微微下降|微降|微微走低|小幅走低|稍稍下跌|小幅走跌
下跌|下降|走低|走弱|走跌
明顯向下走低|清楚顯現跌勢|明白顯露疲弱態勢
重跌下挫|跳樓暴跌|重挫走跌|跳水大跌|深跌入谷|大幅走跌下挫
		"""
		wordLst = xstr.strip().splitlines()
	elif wordLst is None:
		wordLst = ['advanced', 'gained', 'increased', 'rose', 'was quite flat', 'slipped', 'decreased', 'lose', 'dropped']
	if ascendingTF is True:
		idx = loc_aindex(r,thd=thd,openTF=openTF,sortTF=sortTF)
	else:
		idx = loc_dindex(r,thd=thd,openTF=openTF,sortTF=sortTF)
	if idx < len(wordLst):
		xlst = (unicode, str) if hasattr(__builtins__,"unicode") else str
		if isinstance(wordLst[idx],xlst):
			xstr =  rdWord(wordLst[idx])
		else:
			xstr = wordLst[idx]
	else:
		xstr = ''
	return rdWord(xstr)

def performance_verb(r,udfLst):
	n=len(udfLst.thd)
	ul=list(filter(lambda x:udfLst.thd[x]-abs(r)<=0, range(n)))[:1]
	ux = n-1 if len(ul)<1 else ul[0]
	udf_str=udfLst.loc[ux,["u","d","f"]]
	xstr= udfStr(r,udf_str)
	return xstr

def btwDates(a,b,bracket=None,fmt=None,lang="en"):
	if lang.lower() == "cn" :
		return btwDates_cn(a,b,bracket,fmt)
	else :
		return btwDates_en(a,b,bracket,fmt)

def btwDates_cn(a,b,bracket=None,fmt=None):
	if bracket is None :
		bracket=["在于","和","之間"]
	if fmt is None :
		fmt="%Y年%m月"
	s=""
	a=ymd2ym(a,ym=fmt)
	b=ymd2ym(b,ym=fmt)
	return s.join([bracket[0],a,bracket[1],b,bracket[2]])

def btwDates_en(a,b,bracket=None,fmt=None):
	if bracket is None :
		bracket=["between","and"]
	if fmt is None :
		fmt="%Y-%m"
	s=" "
	a=ymd2ym(a,ym=fmt)
	b=ymd2ym(b,ym=fmt)
	return s.join([bracket[0],a,bracket[1],b])

def datetime2hm(dtm,dformat="%H:%M",lang="cn",mp3YN=False):
	if lang!="cn":
		return dtm.strftime(dformat)
	dformat = dformat.replace("%d","%d日").replace("-","")
	if mp3YN == True:
		dformat = dformat.replace(":","")
		dformat = dformat.replace('%H',"%H點")
		dformat = dformat.replace('%M',"%M分") if dtm.minute>0 else  dformat.replace('%M','整') 
		dformat = dformat.replace('%S',"%S秒") if dtm.second>0 else  dformat.replace('%S','')
	return dtm.strftime(dformat)

def ymd2datetime(s,dformat="%Y%m%d"):
	if (isinstance(s,(np.integer,int,float)) or s.isdigit() ):
		s = int(s)
		if log(s,10) >= 12:
			s = s/1000
		if log(s,10) >= 9:
			return datetime.datetime.fromtimestamp(s)
	return datetime.datetime.strptime(str(s),dformat)

def ymd2ym(s,ym=None,ymd='%Y-%m-%d',lang="cn"):
	if ym is None:
		ym = "%Y年%m月" if lang=="cn"  else "%Y-%m"
	return ymd2datetime(s,ymd).strftime(ym)

def ymd2ymd(s,ym=None,ymd='%Y-%m-%d',lang="cn",):
	if ym is None:
		ym = "%Y年%m月%d日" if lang=="cn"  else "%Y-%m-%d"
	if isinstance(s, datetime.date):
		return s.strftime(ym)
	else:
		return ymd2datetime(s,ymd).strftime(ym)

def ymd2md(s,ym='%m-%d',ymd='%Y%m%d',lang="en",mp3YN=False):
	if isinstance(s, datetime.date):
		dtm = s
	else:
		dtm = ymd2datetime(s,ymd)
	if lang=="cn":
		ym = ym.replace("%d","%d日").replace("-","")
		vdu = list(filter(lambda x: x in ym, 'mbB'))
		if len(vdu)>0:
			xdu = '%'+vdu[0]
			ym = ym.replace(xdu,"%m月")
		if mp3YN == True:
			ym = ym.replace(":","")
			ym = ym.replace('%H',"%H點")
			ym = ym.replace('%M',"%M分") if dtm.minute>0 else  ym.replace('%M','整') 
			ym = ym.replace('%S',"%S秒") if dtm.second>0 else  ym.replace('%S','')
	else:
		locale.setlocale(locale.LC_TIME,"en_US.UTF8")
		if ym is None or len(ym)<4:
			ym='%m-%d'
	return dtm.strftime(ym)

def pattern_cn(name,lang="cn"):
	ptnObj={"morning_star":"晨星","evening_star":"夜星",
	"bullish_harami":"曙光乍現","bearish_harami":"烏雲蓋頂",
	"3_white_soldiers":"紅三兵","3_black_crows":"黑三兵",
	"MACD":"MACD看漲" , "bearish_MACD":"MACD看跌" }
	if lang.lower() == "cn" and name in ptnObj:
		return ptnObj[name]
	else :
		return name

def fq2unit_str(fqStr,lang="cn"):
	""" return unitStr from fqStr
	"""
	if lang.lower() == "cn" :
		unitObj={"D":"個交易日","W":"週","M":"個月","Q":"季","Y":"年"}
	else :
		unitObj={"D":"day","W":"week","M":"month","Q":"quarter","Y":"year"}
	return unitObj[fqStr] if fqStr.upper() in unitObj else unitObj["D"]

def rdWord(wordLst,sep="|"):
	""" Select random word from a word list separated with "|" pipe delimeter 
		Args:
			wordLst: list of words glue with "|"
			sep: word separator 
		Returns:
			random selected word
	"""
	xu=wordLst.split(sep)
	return(random.choice(xu))

def udfSet(udf=None,lang="en",tense="past"):
	if lang == "en":
		udx=["increased","decreased","been flat"] if tense=="past" else ["increases","decreases","is basically flat"]
	else:
		udx=["上升|攀升|上漲","下跌|下降","上下稍稍波動|上下輕微震盪|區間微微震盪"]
	if udf is None:
		return udx
	elif type(udf) == str:
		udf = [udf]
	ns=min(len(udf),len(udx))
	for j in range(ns):
		udx[j]=udf[j]
	return udx 

#def udf_word(r,udf=None,zs=[0,0],lang="en",tense="past"):
#	return udfStr(r,udf=udf,zs=zs,lang=lang,tense=tense)

# ['\u2191','\u2193','']

def udfSym(r,udf=[u'\u25B2',u'\u25BC',''],zs=0):
	if  (r>zs):
		xu=udf[0]
	elif(r<zs):
		xu=udf[1]
	else:
		xu=udf[2]
	return xu

def udfStr(r,udf=None,zs=[0,0],lang="en",tense="past"):
	""" 
	Directional descriptor based on value: r 
		r>zs[0], within [ zs[0],zs[1] ], zs[1]>r
		Args:
			r: float or int value
			udf: a list of 3 words 
			zs: threshold values 
		Returns:
			Directional verb/adjective
	"""
	try:
		udx=udfSet(udf,lang=lang,tense=tense)
		if isinstance(zs,(np.integer,int,float)) is True:
			z = abs(zs)
			azs=[z,-z]
		elif len(zs)<1: 
			azs=[0,0]
		elif len(zs)<2: 
			z = abs(zs)
			azs=[z,-z]
		else:
			azs=zs
		if  (r>azs[0]) :
			xu=udx[0]
		elif(r<azs[1]) :
			xu=udx[1]
		else :
			xu=udx[2]
		return(rdWord(xu))
	except Exception as e:
		return(str(e))
	
def udfStr_DEPRECATED(r,udf=None,zs=0,lang="en",tense="past"):
	""" DEPRECATED
	Directional descriptor based on value:r  [r>zs,r<-zs,in-between]
		Args:
			r: float or int value
			udf: a list of 3 words 
			zs: threadhold value
		Returns:
			Directional verb/adjective
	"""
	try:
		udx=udfSet(udf,lang=lang,tense=tense)
		azs=abs(zs)
		if  (r>azs) :
			xu=udx[0]
		elif(r<-azs) :
			xu=udx[1]
		else :
			xu=udx[2]
		return(rdWord(xu))
	except Exception as e:
		return(str(e))
	
def roundMP3(x,n=0,usdSign=None,dotSign='.',lang=None):
	""" Audio format expression
		Args:
			x: Float or int value
			n: Rounding decimal
			usdSign: not used, only for being compatible with ruondUSD
			dotSign: 'point' for reading purpose
	"""
	if n<=0:
		return '{:.0f}'.format(x)
	else:
		ri,rj = '{:.{}f}'.format(x,n).split(".")
		rmd = dotSign + ' '.join(rj) if int(rj)>0 else ''
		return '{}{}'.format(ri,rmd)

def roundAuto(x,n=2,usdSign=None,dotSign='.',lang='en',mp3YN=False):
	if isinstance(x,(list,np.ndarray)):
		return [roundUSD(d,n=n,usdSign=usdSign,dotSign=dotSign,lang=lang,mp3YN=mp3YN) for d in x]
	return roundUSD(x,n=n,usdSign=usdSign,dotSign=dotSign,lang=lang,mp3YN=mp3YN)
	
def roundUSD(x,n=0,usdSign='$',dotSign='.',lang='en',mp3YN=False):
	""" Accounting format expression
		Args:
			x: Float or int value
			n: Rounding decimal
			usdSign: Dollar sign
		Returns:
			Accounting-format string
		Note:
			show simple rounded number if usdSign is None
	"""
	if mp3YN is True:
		xstr = roundSD(x,n=n,lang=lang,mp3YN=mp3YN,TBMTF=True)
		if lang=='cn':
			usdSign='美元'
			xstr = "{}{}".format(xstr,usdSign)
		else:
			xstr = "{}{}".format(xstr,usdSign)
		#return roundMP3(x,n=n,usdSign=usdSign,dotSign=dotSign,lang=lang)
		return xstr
	elif float(x)>10**6:
		xstr = roundSD(float(x),n=n,lang=lang,mp3YN=mp3YN,TBMTF=True)
		xf = usdSign + num_en2cn(xstr) if lang=='cn' else xstr
		return xf
	if n is None: # single digit or less round to 1 decimal else round to integer
		n=0 if abs(x)>=10 or abs(x)%1<0.05 else 1
	else: # if rounding digits are 0s do not show
		n=0 if abs(x)%1<0.5/10**int(n) else n
	if usdSign is None:
		if n<0: # use significant precision if precision n is negative
			xf='{:.g}'.format(x)
		else:
			xf='{:.{}f}'.format(x,n)
	else:
		xf='{}{:0,.{}f}'.format(usdSign,x,n)
	return(xf)

def wrapWord(a,bk=["(",")"]):
	return '{1}{0}{2}'.format(*([a]+bk))

def rangeStr(a,b,bk=["[","]"]):
	return '{2}{0}, {1}{3}'.format(*([a,b]+bk))

def btwStr(a,b,bk=["ranging between","and"],fl="fluctuated around"):
	if a==b:
		return "{} {}".format(fl,a)
	return '{2} {0} {3} {1}'.format(*([a,b]+bk))

def btwStr_cn(a,b,bk=["介于","至","的範圍之間"],fl=["變化不大，只會在","內上下稍稍波動"],unitStr=''):
	if a==b:
		return "{2} {0}{1}{3}".format(*([a,unitStr]+fl))
	return '{2} {0} {3} {1} {5}{4}'.format(*([a,b]+bk+[unitStr]))

def connMdb(clientM=None):
	if isinstance(clientM,MongoClient):
		return clientM
	elif clientM is None: # as None
		uriM = "localhost:27017"
		clientM=MongoClient(uriM)
	elif isinstance(clientM,str): # as hostname
		uriM = "{}:27017".format(clientM)
		clientM=MongoClient(uriM) # existed connection 
	else:
		return clientM
	return clientM

def find_mdb(jobj={},clientM=None,dbname='ara',tablename=None,field={},sortLst=[],ascendingTF=False,limit=0,dfTF=False,**optx):
	""" find WHERE: jobj from mongodb [dbname]::[tablename]
	 and display FIELD: field
		Args:
			a json object
		returns:
			jobj if success
	"""
	#from bson.json_util import loads,dumps
	mobj=[]
	try:
		clientM=connMdb(clientM)
		if tablename is None:
			return jobj,clientM,"tablename not defined!"
		if hasattr(clientM,'_Database__client') and isinstance(clientM._Database__client,MongoClient):
			dbM=clientM
		else:
			dbM=clientM[dbname]
		zobj={k:1 for k in field } if len(field)>0 else {}
		if len(sortLst)>0 and isinstance(sortLst,dict):
			# Note, sorting might be diff as dict variables are not in order
			zsrt=[(k,v) for k,v in sortLst.items()]
		elif len(sortLst)>0 and isinstance(list(sortLst)[0],(tuple,list)):
			zsrt=sortLst
		elif ascendingTF:
			zsrt=[(k,1) for k in sortLst] if len(sortLst)>0 else [('_id',1)]
		else:
			zsrt=[(k,-1) for k in sortLst] if len(sortLst)>0 else [('_id',-1)]
		if '_id' not in field:
			zobj.update(_id=0)
		#zobj=loads(dumps(zobj))
		if limit>0:
			mbson = dbM[tablename].find(jobj,zobj,sort=zsrt).limit(limit)
		else:
			mbson = dbM[tablename].find(jobj,zobj,sort=zsrt)
		#mobj= json.loads(dumps(mbson))
		mobj= [x for x in mbson]
		clientM.close()
		msg="success"
	except Exception as e:
		msg=str(e)
		sys.stderr.write("**ERROR: {} @{}.find()\n".format(str(e),tablename))
	if len(mobj)>0 and len(mobj[0])<1:
		mobj=[]
	if dfTF:
		mobj = pd.DataFrame(mobj)
	return mobj,clientM,msg

def get2mndb(jobj,tablename='lsi2nlg',zpk={'ticker','username','tmplname','category','tmpltype','lang'}):
	""" find jobj from mongodb [ara:ls2nlg] based on 'username' & 'tmplname' keys 
		Args:
			jobj:json object
			zpk: set of primary keys 
		returns:
			jobj with 'retcode' updated
	"""
	try:
		uriM = "localhost:27017"
		clientM=MongoClient(uriM)
		dbM=clientM['ara']
		
		zobj={k: jobj[k] for k in set(jobj.keys()) & zpk } if len(zpk)>0 else jobj
		# find via 'ticker','username','tmplname','category','tmpltype','lang'
		mobj=dbM[tablename].find_one(zobj)
		# find via 'username','tmplname','category','tmpltype','lang'
		if mobj is None:
			if 'ticker' in zobj:
				del zobj['ticker']
			mobj=dbM[tablename].find_one(zobj)
		# find via 'username','category','tmpltype','lang'
		if mobj is None:
			zobj['username'] = 'system'
			if 'tmplname' in zobj:
				del zobj['tmplname']
			mobj=dbM[tablename].find_one(zobj)
		if mobj is None:
			jobj['retcode']='No record found on {}'.format(zobj.items())
			return jobj
		else: 
			mobj['username'] = jobj['username'] if 'username' in jobj else 'system'
		if len(jobj)>1:
			for j,xky in enumerate(jobj):
				if xky not in mobj:
					continue
				jobj[xky]= mobj[xky]
		else:
			jobj.update(mobj)
		clientM.close()
		if 'tmplrpt' not in jobj:
			jobj['tmplrpt']= ''
		if 'lastmod' in mobj:
			jobj['retcode'] += " Last mod. @ {}".format(mobj['lastmod'])
		else:
			jobj['retcode']='Record found via {}'.format(zobj.items())
		sys.stderr.write("====={}\n".format(jobj))
	except Exception as e:
		jobj['retcode']='**ERROR: {} @ get2mndb()'.format(str(e))
	return jobj

@safeRunArg({})
def upsert_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,wmode='upsert',zpk=[],zasc=[],**optx):
	"""
	upsert 'jobj' to MDB 'dbname'::'tablename'
	  w.r.t unique primary keys 'zpk' defined as zip('zpk','zasc')
	Optional parameters: 
	  pkTF=True  : To create primary keys
	  uniqueTF=True : To ensure unique primary key
	  ordered=False : to ensure insert even duplicate errors occur 
	Example:
	upsert_mdb({"name":"Mary"},tablename="test",zpk=["name"])
	upsert_mdb({"name":"Mary"},tablename="test",zpk=["name"],pkTF=False)
	"""
	return insert_mdb(jobj=jobj,clientM=clientM,port=port,dbname=dbname,tablename=tablename,wmode=wmode,zpk=zpk,zasc=zasc,**optx)

# upsert_mdb wrapper
def write_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,wmode='upsert',zpk=[],zasc=[],**optx):
	return insert_mdb(jobj=jobj,clientM=clientM,port=port,dbname=dbname,tablename=tablename,wmode=wmode,zpk=zpk,zasc=zasc,**optx)

@saferun
def delete_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,**optx):
	clientM=connMdb(clientM)
	ret=clientM[dbname][tablename].delete_many(jobj)
	return ret

def insert_mdb(jobj={},clientM=None,port=27017,dbname='ara',tablename=None,wmode='',zpk=[],zasc=[],**optx):
	"""
	Insert 'jobj' to MDB 'dbname'::'tablename'
	  w.r.t unique primary keys 'zpk' defined as zip('zpk','zasc')
	Optional parameters: 
	  pkTF=True  : To create primary keys
	  uniqueTF=True : To ensure unique primary key
	  ordered=False : to ensure insert even duplicate errors occur 
	Example:
	insert_mdb({"name":"Mary"},tablename="test",zpk=["name"])
	insert_mdb({"name":"Mary"},tablename="test",zpk=["name"],pkTF=False)
	"""
	if not all([len(jobj),tablename]):
		errmsg="**ERROR:{} @{}\n".format(tablename,"insert_mdb()")
		return jobj,None,errmsg
	try:
		pkTF=getKeyVal(optx,'pkTF',True)
		uniqueTF=getKeyVal(optx,'uniqueTF',True)
		ordered=getKeyVal(optx,'ordered',False)
		debugTF=getKeyVal(optx,'debugTF',False)
		createTF=getKeyVal(optx,'createTF',False)
		clientM=connMdb(clientM)
		if hasattr(clientM,'_Database__client') and isinstance(clientM._Database__client,MongoClient):
			mdb=clientM
		else:
			mdb=clientM[dbname]
		try:
			if createTF is True or wmode in ['create','replace']:
				ynX=mdb[tablename].drop()
				sys.stderr.write("Drop collection:{}, status:{}\n".format(tablename,ynX))
			if len(zpk)>0 and pkTF is True:
				if any( [ len(zasc)<1, len(zasc)!=len(zpk) ]):
					zasc=[1]*len(zpk)
				pkLst = [(k,v) for k,v in zip(zpk,zasc)]
				mdb[tablename].drop_indexes()
				mdb[tablename].create_index(pkLst,unique=uniqueTF)
				if debugTF:
					emsg="\n==INDEX pkLst:{},zpk:{},zasc:{} on {},optx:{}\n".format(pkLst,zpk,zasc,tablename,optx)
					sys.stderr.write(emsg)
		except Exception as e:
			errmsg="==ERROR:{} @{}\n".format(str(e),"insert_mdb:create_index")
			sys.stderr.write(errmsg)
			pass
		if isinstance(jobj,pd.DataFrame):
			jobj = jobj.to_dict(orient='records')
		elif isinstance(jobj,dict):
			jobj = [jobj]
		if getKeyVal(optx,'jsonObjReloadTF',False):
			jobj = jsonObjReload(jobj)
			sys.stderr.write(" --{} @ {}\n".format('jsonObjReload','insert_mdb()'))
		if wmode=='upsert':
			ret=upd2mdb(jobj,zpk=zpk,mcur=mdb[tablename])
		else:
			ret=mdb[tablename].insert_many(jobj,ordered=ordered)
		errmsg="Success:{}".format(len(ret))
	except Exception as e:
		errmsg="**ERROR:{} @{}\n".format(str(e),"insert_mdb:insert_many")
		sys.stderr.write(errmsg)
		try:
			sys.stderr.write("---JOBJ[0]:\n{}\n".format(jobj[0]))
		except:
			pass
		return {},None,errmsg
	return jobj,mdb,errmsg

def save2mgdb(mLst=None,mky=None,dbM=None,tablename="temp",wmode='replace'):
	if dbM is None or mLst is None:
		return None
	createTF=True if wmode=='replace' else False
	dy,_,errmsg=insert_mdb(jobj=mLst,zpk=mky,dbM=dbM,tablename=tablename,createTF=createTF)
	return len(dy)

def upd2many(dd=[],zpk={},mcur={},debugTF=False):
	''' update multiple records based on each `zpk` key set
		Note, apply with caution as if zpk is not unique index
		may cause undesired results  
	'''
	if debugTF:
		sys.stderr.write(" --update via `update_many` @{}\n".format("upd2many()"))
	for j, mobj in enumerate(dd):
		jobj = {k:v for k,v in mobj.items() if k in zpk}
		r=mcur.update_many(jobj,{"$set":mobj},upsert=True)
		#print(' ---{}.{}:{}'.format(j,mobj,r.raw_result))
	return dd
	
def upd2mdb(dd=[],zpk={},mcur={},debugTF=False):
	''' support function to insert_mdb as upsert_mdb
		maintain _id if exists or add one for new record
	'''
	if debugTF:
		sys.stderr.write(" --update via `replace_one` @{}\n".format("upd2mdb()"))
	for j, mobj in enumerate(dd):
		jobj = {k:v for k,v in mobj.items() if k in zpk}
		r = mcur.replace_one(jobj,mobj,upsert=True)
		#print(' ---{}.{}:{}'.format(j,mobj,r.raw_result))
	return dd
	
def upd2mdbOLD(dd=[],zpk={},mcur={},debugTF=False):
	''' support function to insert_mdb
	mongodb utility of insert_mdb/upsert_mdb
	remove and insert for update based on zpk primary key
	'''
	def remove_doc(jobj,zpk,nz):
		ret={}
		if nz<1:
			return ret
		jx = { ky:va for ky,va in jobj.items() if ky in zpk }
		if len(jx)==nz:
			if debugTF:
				sys.stderr.write(" --remove primany key doc:{}\n".format(jx))
			ret = mcur.remove(jx)
		elif len(jx)<nz:
			sys.stderr.write(" --{} missing primary keys:{}\n".format(jx,zpk))
		return ret

	nz = len(zpk)
	for j, jobj in enumerate(dd):
		try:
			if debugTF:
				sys.stderr.write(" --insert doc:{}:{}:{}\n{}\n".format(type(jobj),jobj.keys(),mcur,jobj))
			jobj.pop('_id',None)
			remove_doc(jobj,zpk,nz)
			ret = mcur.insert(jobj)
		except Exception as e:
			errmsg="**ERROR:{}:{} @{}\n{}\n".format(j,str(e),"upd2mdb()",jobj)
			sys.stderr.write(errmsg)
	return dd

def np2mongo(zobj):
	for j,(x,y) in enumerate(zobj.items()):
		if isinstance(y,np.integer):
			zobj[x] = int(y)
		elif isinstance(y,np.float):
			zobj[x] = float(y)
	return zobj

def write2mdb(jobj,clientM=None,dbname='ara',tablename=None,zpk={},insertOnly=False):
	"""
	write & replace jobj to mongodb [dbname]::[tablename]
	based on primary keys in {} dict format: {zpk}
	Args:
		jobj: a dataFrame, list_of_object or object
	returns:
		list_of_object, mongo+database_handler, error_msg
	Note: 
	  1. zpk={'*'}:  delete to any existed records before insertion 
	  2. insertOnly=True: insert record without checking any duplication
	  3. 'jobj' return with additional '_id' field after execution
	"""
	def jobj2mdbc(jobj,mdbc,zpk,insertOnly=False):
		mobj = jobj # jobj.copy()
		if mobj.get('_id',False):
			del mobj['_id']
		if insertOnly:
			mdbc.insert(mobj)
			return mobj
		try:
			if len(zpk)<1:
				xpk = {}
			else:
				xpk = set(mobj.keys()) & set(zpk)
			if len(xpk)>0 and len(xpk)>=len(set(zpk)):
				zobj={k: mobj[k] for k in xpk }
				try:
					zobj = np2mongo(zobj)
					mdbc.delete_many(zobj)
				except Exception as e:
					sys.stderr.write("**ERROR: {} @ {}, mdbc:{}, zpk:{}, zobj:\n{}\n".format(str(e),'delete_many()',mdbc,zpk,zobj) )
					sys.stderr.write("===rpt_hm: type:{},val:{}".format(type(zobj['rpt_hm']),zobj['rpt_hm']))
		except Exception as e:
			sys.stderr.write("**ERROR: {} @ {}, zpk:{}, jobj:\n{}\n".format(str(e),'jobj2mdbc()',zpk,jobj) )
		mobj = np2mongo(mobj)
		mdbc.insert(mobj)
		return mobj

	if clientM is None:
		uriM = "localhost:27017"
		dbM=MongoClient(uriM)[dbname]
	elif isinstance(clientM,str) is True:
		uriM = "{}:27017".format(clientM)
		dbM=MongoClient(uriM)[dbname]
	elif isinstance(clientM,MongoClient):
		dbM=clientM[dbname]
	elif hasattr(clientM,'_Database__client') and isinstance(clientM._Database__client,MongoClient):
		dbM = clientM
	elif isinstance(clientM,MongoClient) is False:
		return jobj,None,"no mongodb driver"
	mobj={}
	try:
		msg="success"
		if isinstance(jobj,list) is True:
			for xobj in jobj:
				try:
					mobj = jobj2mdbc(xobj,dbM[tablename],zpk=zpk,insertOnly=insertOnly)
				except Exception as e:
					sys.stderr.write("**ERROR: {} @ {}\n".format(str(e),'write2mdb() List-Dict') )
					msg=str(e)
					continue
		elif isinstance(jobj,pd.DataFrame) is True:
			df = jobj
			jobj = df.to_dict(orient='records')
			if len(zpk)>0 and set(zpk) == {'*'}:
				dbM[tablename].delete_many({})
				zpk = {}
			for xobj in jobj:
				try:
					mobj = jobj2mdbc(xobj,dbM[tablename],zpk=zpk,insertOnly=insertOnly)
				except Exception as e:
					sys.stderr.write("**ERROR: {} @ {}\n".format(str(e),'write2mdb() DataFrame') )
					msg=str(e)
					continue
		else:
			try:
				mobj = jobj2mdbc(jobj,dbM[tablename],zpk=zpk,insertOnly=insertOnly)
			except Exception as e:
				sys.stderr.write("**ERROR: {} @ {}, zpk:{}, jobj:\n{}\n".format(str(e),'write2mdb() Dict',zpk,jobj) )
				msg=str(e)
		#clientM.close()
	except Exception as e:
		#import pickle
		msg=str(e)
		sys.stderr.write("**ERROR: {} @ {}\n".format(msg,'write2mdb()') )
		#fp = open("mdb_err.pickle","wb")
		#pickle.dump(jobj,fp)
		#fp.close()
	return mobj,dbM,msg

def df_tofile(fp,df,output=None,sep='|',indexTF=False,hdrTF=True):
	"""
	Write dataframe:df to file via file handle:fp
	where
		fp: file handle
		df: data in dataframe
		output: output format of [csv|json|html|string|None], optional
		sep: output separator, default to '|', optional
		indexTF: flag to show index, default to False, optional
		hdrTF: flag to show header, default to True, optional
	"""
	ret = df_output(df,output=output,sep=sep,indexTF=indexTF,hdrTF=hdrTF)
	str_tofile(fp,ret)

def str_tofile(fp,s):
	"""
	Write string:s to file via file handle:fp
	"""
	try:
		if isinstance(fp,str) is True:
			fp = sys.stdout if fp =='-' else open(fp,'w') 
		fp.write(s)
		ret = 1
	except Exception as e:
		sys.stderr.write("**ERROR: {} @ {}\n".format(str(e),'str_tofile()') )
		ret = 0
	return ret

def df_output(df,output=None,sep='|',indexTF=False,hdrTF=True):
	"""
	convert dataframe:df to a string 
	where
		df: data in dataframe
		output: output format of [csv|json|html|string|None], optional
		sep: output separator, default to '|', optional
		indexTF: flag to show index, default to False, optional
		hdrTF: flag to show header, default to True, optional
	"""
	if df is None or len(df)<1:
		return ''
	if output == 'csv':
		ret = df.to_csv(sep=sep,index=indexTF,header=hdrTF)
	elif output == 'json':
		ret = df.to_json(orient='records')
	elif output == 'html':
		ret = df.to_html(index=indexTF,header=hdrTF)
	elif output == 'string':
		ret = df.to_string(index=indexTF,header=hdrTF)
	return ret

def save2mndb(jobj,zpk={'ticker','username','tmplname','category','tmpltype','lang'},clientM=None,dbname='ara'):
	""" 
	save jobj into mongodb [ara:ls2nlg] based on 'username' & 'tmplname' keys 
	Args:
		jobj: json object
		zpk: list of primary keys if exist
		clientM: mongo driver
		dbname: database name default to 'ara'
	returns:
		jobj with 'retcode' & 'lastmod' updated
	Note, create new table if zpk is {'*'} 
	"""
	try:
		mobj=jobj.copy()
		if jobj.get('retcode',False):
			del mobj["retcode"]
		del mobj["action"]
		if clientM is None:
			uriM = "localhost:27017"
			clientM=MongoClient(uriM)
		dbM=clientM[dbname]
		if set(zpk) == {'*'}:
			zobj = {}
			xpk = {}
		else:
			xpk = set(mobj.keys()) & set(zpk)
		if len(xpk)>0:
			zobj={k: mobj[k] for k in xpk }
		dbM.lsi2nlg.delete_many(zobj)
		dbM.lsi2nlg.insert(mobj)
		clientM.close()
		jobj["retcode"]="save successfully"
	except Exception as e:
		jobj["retcode"]="**ERROR:{} @ save2mndb()".format(e)
	return jobj

def lst2dict(jkst,jlst):
	""" combine 2 lists of shorter one to be a (key,value) object 
		Args:
			jkst: key list
			jlst: value list
		Returns:
			a (key,value) object
	"""
	jobj={}
	for j in range( min(len(jkst),len(jlst))):
		jobj[jkst[j]]=jlst[j]
	return jobj

def tmpl2lsi(tmplstr,argobj):
	""" Literal String Interpolation to format a text string
		Args:
			tmplstr: template text string
			argobj: a dictionary object 
		Returns:
			formatted interpolated text string
	"""
	if not argobj:
		return tmplstr
	else:
		return jj_fmt(tmplstr,argobj)

def combine_cmd2dict(*_xargs_): # TBD ,**_xdu_):
	""" combine list of scripts [_xargs_] and exec them  
		return result as dict
		Note: _xargs_, _xstr_ are reserved variables
		Also see: qs_exec()
	"""
	try:
		_xstr_="\n".join(_xargs_)
		vux = qs_exec(_xstr_)
	except Exception as e:
		sys.stderr.write("**ERROR: {} @{}\n".format(str(e),"combine_cmd2dict"))
		return {}
	dux={}
	for x,v in vux.items():
		if isinstance(v, types.FunctionType) is True:
			continue
		elif x in ['_xargs_','_xstr_','_xdu_']:
			continue
		else:
			dux[x] = v
	sys.stderr.write("==Variables @{}():{}\n".format("combine_cmd2dict",dux))
	return dux

def jobj2lsi(jobj):
	""" Literal String Interpolation to format a text string
		Args:
			jobj: a dictionary object that contains 8 elements with keys:
			["tmplstr","argstr","prepstr","tmplrpt","username","category",
			"tmplname","action"]
			for interpolation argument list and pre-process
		Returns:
			formatted interpolated text string
	"""
	tmplrpt=""
	try:
		# combine script prepstr,argstr and exec them  
		if "prepstr" not in jobj:
			jobj["prepstr"]=''
		jobj["argobj"]=combine_cmd2dict(jobj["prepstr"],jobj["argstr"])
	except Exception as e:
		retcode="**ERROR: {} @{}() ".format(str(e),"combine_cmd2dict")
		return (tmplrpt,retcode)

	#- Run lsi to generate context
	try:
		tmplrpt=tmpl2lsi(jobj["tmplstr"],jobj["argobj"])
		retcode="run successfully"
	except Exception as e:
		retcode="**ERROR: {} @ tmpl2lsi() ".format(e)
	return (tmplrpt,retcode)

def lsi2nlg_calc(jobj):
	jobj["lastmod"]=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	#if jobj.keys()<8 :
	#	return jobj
	#elif jobj["action"] == 'save':
	if jobj["action"] == 'save':
		jobj=save2mndb(jobj)
	elif jobj["action"] == 'get':
		jobj=get2mndb(jobj)
	elif jobj["action"] == 'run':
		(jobj["tmplrpt"],jobj["retcode"])=jobj2lsi(jobj)
	return jobj

def qs_split(xstr,d1='&',d2='='):
	""" split query string into a dict object """
	d = {k:v for k,v in [lx.split(d2) for lx in xstr.split(d1)]}
	return d

def qs_exec(xstr):
	""" convert commend string into a dict object """
	d = {}
	exec(xstr,globals(),d)
	return d

def qs_execOLD(xstr):
	""" convert commend string into a dict object """
	exec(xstr)
	del xstr
	d = locals()
	return d

def extra_opts(opts={},xkey='',method='JS',updTF=True):
	""" 
	update values of extraKey: xkey to DICT: opts 
	based on method: [JS|QS|XS]
	where
		JS: DICT/JSON type object
		QS: k1=v1;k2=v2;... query-like string
	"""	
	import ast	
	if xkey is None or len(xkey)<2:
		xkey = "extra{}".format(method)
	if xkey not in opts or opts[xkey] is None: 
		d = {}
	elif method == 'JS': # additional key/val like POST arguments
		d = ast.literal_eval(opts[xkey])
	elif method == 'QS': # additional key/val like GET arguments
		d = qs_split(opts[xkey])
	else: # additional key/val like python executable string
		d = qs_exec(opts[xkey])
	if updTF is True:
		opts.update(d)
		opts[xkey] = None
	return d

def generate_comment_hourly(ts,df,opts=None):
	import types
	import ast
	from _alan_calc import subDict
	from _alan_pppscf import vertex_locator,vertex_mnmx
	if opts is None and 'opt_csv2plot' in globals():
		opts, _ = opt_csv2plot([])
	if 'open' not in df and 'close' in df:
		df['open'] = df['close'].copy()
	if 'volume' not in df:
		df['volume'] = 0.0
	saveDB=getKeyVal(opts,'saveDB',True)
	debugTF=getKeyVal(opts,'debugTF',False)
	intradayTF=getKeyVal(opts,'intradayTF',False)
	dd = subDict(opts,['j2ts','j2name'],reverseTF=True)
	#if opts['extraJS'] is not None:
	#	try:
	#		dd.update(ast.literal_eval(opts['extraJS']))
	#	except Exception as e:
	#		print >> sys.stderr, str(e)
	if debugTF is True:
		sys.stderr.write("opts in generate_comment_hourly():\n{}\n".format(opts))
	if 'volstats' in dd and dd['volstats'] is not None:
		xmn,xsd = [float(x) for x in dd['volstats'].split(',') ]
		ubound,lbound = (xmn+xsd/2.0, xmn-xsd/2.0)
	else:
		ubound,lbound = 0,0
	dd.update(ubound=ubound,lbound=lbound)
	if 'rpt_hm' in dd:
		endTF = True if int(dd['rpt_hm'])>1530 else False
	else:
		endTF = False
	if df.size>0 and ts:
		dd.update(f=df)
		if 'close' in df:
			volS=dd['cvol'] if 'cvol' in dd else df['volume'].sum()*390/df.shape[0]
			dg, dh = vertex_locator(df['close'],**subDict(opts,['npar','debugTF']))
			#dh.reset_index(drop=True,inplace=True)
			if debugTF is True:
				sys.stderr.write("dh:\n{}\n".format(dh))
			prc_open = df['open'].iloc[0] if intradayTF is True else df['close'].iloc[0]
			# re-adjust cvol during the trading hours
			if endTF is True and 'cvol' in dd: 
				volS=dd['cvol']
			elif intradayTF is True: 
				volS = df['volume'].sum()*390/df.shape[0] 
			else:
				volS = df['volume'].iloc[-1]
			dk = vertex_mnmx(dh,prc_open,colx='actual')
			if debugTF is True:
				sys.stderr.write( "dMmMx:\n{}\n".format(dk))
			dd.update(dfdr=dh,dMnMx=dk,volS=volS,copen=prc_open)
			dd.update(fcsTrend=dh['sign'].iloc[-1]) # forthcoming trend 1:up,-1:down, 0: fair
		if debugTF is True:
			sys.stderr.write("==dd:\n{}\n".format(dd))
			sys.stderr.write("==Template ts:\n{}\n".format(ts))
		ret =  jj_fmt(ts,**dd)
		if debugTF is True:
			sys.stderr.write("==Comment:\n{}\n".format(ret))
		fname="hourly_headline_cn.j2"
		hlts = '{} include "{}" {}'.format('{%',fname,'%}')
		hlret =  jj_fmt(hlts,**dd)
		if debugTF is True:
			sys.stderr.write("==headline:\n{}\n".format(hlret))
		dd.update(mp3YN=True)
		mp3ret =  jj_fmt(ts,**dd)
		if debugTF is True:
			sys.stderr.write("==mp3Comment:\n{}\n".format(mp3ret))
		# FOR Technical Investment Strategy RUN
		techTF = getKeyVal(dd,'techTF',False)
		#if 'rpt_status' in dd and dd['rpt_status'] == 'trading': 
		if techTF is True:
			try:
				#pgDB =  conn2pgdb()
				from ohlc_minute_tst import find_stgy,run_narratives
				from _alan_calc import conn2pgdb,run_tech
				lang = dd['lang']
				if 'clop' not in df.columns:
					ma1=5;ma2=30
					mdf = run_tech(df, pcol='close',winLst=[ma1,ma2],debugTF=debugTF,nanTF=True)
				else:
					mdf=df.dropna(axis=0, how='any')
				minute_pnl = find_stgy(mdf,debugTF=debugTF)
				#retAddi = run_narratives_tst(dd['ticker'],mdf,minute_pnl,debugTF=debugTF)
				retAddi = run_narratives(dd['ticker'],mdf,minute_pnl,lang=lang,debugTF=debugTF,optx=dd)
				if debugTF is True:
					sys.stderr.write("==addiComment:\n{}\n".format(retAddi))
				ret += retAddi
				mp3ret += retAddi
				sys.stderr.write("==minute_pnl:\n{}\n".format(minute_pnl.tail().to_csv()))
			except Exception as e:
				sys.stderr.write( "**ERROR:{} @run_narratives :".format(str(e)) )
				sys.stderr.write( "ticker:{}\n".format(dd['ticker']) )
				sys.stderr.write( "mdf:\n{}\n".format(mdf.head().to_string()))
		tpLst = (types.BuiltinFunctionType,types.FunctionType,pd.DataFrame)
		jobj = { k:v for (k,v) in dd.items() if isinstance(v,tpLst) is False }
		clientM=None
		if opts['src'] in ['tw','hk','cn']:
			mdbname='ara_{}'.format(opts['src'].lower())
		else:
			mdbname='ara'

		jobj['comment'] = ret
		jobj['headline'] = hlret
		jobj['mp3comment'] = mp3ret
		sys.stderr.write("==ret:\n{}==hlret:\n{}\n".format( ret,hlret))
		if saveDB is True:
			for (k,v) in dd.items():
				if isinstance(v,pd.DataFrame) is True:
					jobj[k] = v.to_dict(orient='records')
			mobj,clientM,err_msg = write2mdb(jobj,clientM,dbname=mdbname,tablename='hourly_report',zpk={'label','rpt_date','rpt_hm'})
		else:
			for (k,v) in dd.items():
				if isinstance(v,pd.DataFrame) is True:
					jobj[k] = v
		if debugTF is True:
			sys.stderr.write( "==jobj in generate_comment_hourly():\n{}\n".format(jobj))
		return jobj
	else:
		return {}

def trend_forecast(ticker='^GSPC',freq='W',debugTF=False,pgDB=None,**optx):
	'''
	forecast daily/weekly/monthly as D/W/M forecasts of 'ticker'
	'''
	try:
		from _alan_ohlc_fcs import run_ohlc_fcs;
		(dd,dwm,datax)=run_ohlc_fcs(ticker,freq=freq,debugTF=debugTF,pgDB=pgDB,**optx)
		dc = dwm.query('freq=="{}"'.format(freq)).to_dict(orient='records')[0]
		dc.update(f=datax)
	except Exception as e:
		dc = dict(ticker=ticker,pos_pb=0,f={})
	return dc

def stock_screener(scrIds='most_actives',sortLst=['dollarValue'],ascLst=[False],**optx):
	'''
	get stock screened by 'scrIds'
	where 'scrIds' can be any of:
	  day_gainers,day_losers,most_actives,most_values
	  sector_gainers,sector_losers or customized 'scrIds'
	and return a dataframe up to 'nobs' rows with columns 'colx' default to:
	ticker, price, volume, change, changePercent, marketCap, dollarValue, sector, sector_cn, company, company_cn, pbdt
	sorted by columns 'sortLst' with ascending order as 'ascLst'
	'''
	scrIds= scrIds.lower()
	if scrIds=='day_gainers':
		df = x_screener(sortLst=['changePercent'],ascLst=[False],**optx)
	elif scrIds=='day_losers':
		df = x_screener(sortLst=['changePercent'],ascLst=[True],**optx)
	elif scrIds=='most_actives':
		df = x_screener(sortLst=['volume'],ascLst=[False],**optx)
	elif scrIds=='most_values':
		df = x_screener(sortLst=['dollarValue'],ascLst=[False],**optx)
	elif scrIds in ['5%','+5%']:
		optx.update(addiFilter='changePercent>=5&price>7.99',nobs=10)
		df = x_screener(sortLst=['changePercent'],ascLst=[False],**optx)
	elif scrIds=='-5%':
		optx.update(addiFilter='changePercent<-5&price>7.99',nobs=10)
		df = x_screener(sortLst=['changePercent'],ascLst=[True],**optx)
	elif scrIds=='sector_gainers':
		addiClause="and ticker in ('XLB','XLE','XLF','XLI','XLK','XLP','XLRE','XLU','XLV','XLY','XTL')"
		optx.update(addiFilter=0,nobs=11)
		df = x_screener(sortLst=['changePercent'],ascLst=[False],addiClause=addiClause,**optx)
	elif scrIds=='sector_losers':
		addiClause="and ticker in ('XLB','XLE','XLF','XLI','XLK','XLP','XLRE','XLU','XLV','XLY','XTL')"
		optx.update(addiFilter=0,nobs=11)
		df = x_screener(sortLst=['changePercent'],ascLst=[True],addiClause=addiClause,**optx)
	else:
		df = x_screener(sortLst=sortLst,ascLst=ascLst,**optx)
	return df

def x_screener(nobs=6,sortLst=['dollarValue'],ascLst=[False],dbname='ara',tablename='yh_qutoe_curr',addiFilter=1,debugTF=False,colx=[],addiClause='',**optx):
	'''
	support function to stock_screener() to get most active stocks 
	from MDB::dbname::tablename
	based on addiFilter criteria
	sorted by columns 'sortLst' with ascending order as 'ascLst'
	return a dataframe up to 'nobs' rows with columns 'colx' default to:
	ticker, price, volume, change, changePercent, marketCap, dollarValue, sector, sector_cn, company, company_cn, pbdt
	'''
	from _alan_calc import sqlQuery,subDF
	da,_,_=find_mdb({},tablename='yh_quote_curr',dbname=dbname,dfTF=True)
	da.dropna(subset=['close','volume'],inplace=True)
	if len(da)<1:
		return {}
	xqTmp="select * from mapping_ticker_cik where act_code=1"
	if len(addiClause)<1:
		addiClause="and sector not like '%%Index'"
	xqr="{} {}".format(xqTmp,addiClause)
	db = sqlQuery(xqr,dbname=dbname)
	da = da.merge(db[['ticker','company_cn','sector','sector_cn']],on='ticker')
	if len(da)<1:
		return {}
	da['price']=da['close']
	da['dollarValue']=(da['volume']*da['close']).astype(int)
	if addiFilter:
		if addiFilter==1:
			addiFilter='abs(changePercent)>0.5&price>7.99'
		elif addiFilter==2:
			addiFilter='abs(changePercent)>5&price>7.99'
		elif addiFilter==3:
			addiFilter='abs(changePercent)>2.5&price>4.99'
		elif isinstance(addiFilter,(int,float)):
			addiFilter='price>4.99'
		if debugTF:
			sys.stderr.write("==addiFilter:{}\n".format(addiFilter))
		try:
			da = da.query(addiFilter)
		except:
			pass
	if len(da)<1:
		return {}
	if len(sortLst)>0:
		df = da.sort_values(by=sortLst,ascending=ascLst)
	else:
		df = da
	df.reset_index(drop=True,inplace=True)
	df['ranking'] = [int(x)+1 for x in df.index.values]
	if colx is None or len(colx)<1:
		colx=['ranking','ticker','price','volume','change','changePercent','marketCap','dollarValue','Range52Week','pbdt','sector','sector_cn','company','company_cn']
	df=subDF(df,colx)
	if nobs>0:
		return df.iloc[:nobs]
	else:
		return df

def ssh2mg(sshUser='rstudio',rhost='api1.beyondbond.com',lhostLst=['bbapi1','api1','api1.beyondbond.com'],port=27017,**optx):
	''' remotely connect to server `rhost` mongoDB via ssh-tunnel
		local connection rhost="localhost" 
	'''
	import subprocess
	lhost=subprocess.Popen("hostname",stdout=subprocess.PIPE).stdout.read().decode().strip()
	rhost= rhost if lhost not in lhostLst else 'localhost'
	return remote2mgdb(sshUser=sshUser,host=rhost,port=port,**optx)

def remote2mgdb(sshUser='rstudio',host='localhost',port=27017,**optx):
	from sshtunnel import SSHTunnelForwarder
	import pymongo
	if not host:
		return None,None
	elif host=='localhost':
		return None, pymongo.MongoClient('localhost', port)
	#ssh_pkey=optx.get('ssh_pkey','/home/'+sshUser+'/.ssh/id_rsa')
	#ssh_pass=optx.get('ssh_pass',open('/home/'+sshUser+'/.sshpass').read().strip())
	ssh_pass=optx.get('ssh_pass','{pass}'.format(**json.load(open('/home/'+sshUser+'/.sshpass.json'))) )
	serverM = SSHTunnelForwarder(host, ssh_username=sshUser,
		#ssh_pkey=ssh_pkey,
		ssh_password=ssh_pass,
		remote_bind_address=('localhost', port))
	serverM.start()
	print("==remote server:",serverM,file=sys.stderr)
	# serverM.local_bind_port is assigned local port
	clientM = pymongo.MongoClient('localhost', serverM.local_bind_port) 
	print("==remote clientM:",clientM,file=sys.stderr)
	#serverM.stop()
	return serverM, clientM

def cgi2dict(**optx):
	'''Parse GET/POST parameters into python dictionary via converting:
		1. any variables end with 'TF' is True/False boolean
		2. any existed `optx` integer elements into "int" type
		3. any "numerical" elements into float respectively
	Note,
		1. cgiTF:True/False will append to `optx`
		2. content_type will also append to `optx` if cgiTF
	'''
	import cgitb,cgi,os
	xreq=os.getenv('REQUEST_METHOD')
	if xreq not in ('GET','POST'):
		optx.update(cgiTF=False)
		return optx
	else:
		optx.update(cgiTF=True)
	logdir_cgi=optx.pop('logdir_cgi','./')
	format_cgi=optx.pop('format_cgi','text')
	display_cgi=optx.pop('display_cgi',0)
	cgitb.enable(display=display_cgi, logdir=logdir_cgi)
	cgitb.enable(format=format_cgi)
	mf = cgi.FieldStorage()
	for ky in mf:
		va=mf[ky].value
		if ky[-2:] == 'TF':
			va = True if va[:1].lower() in ['t','y','1'] else False
		elif ky in optx and isinstance(optx[ky],(int)):
			try:
				va=int(va)
			except:
				pass
		else:
			try:
				va=float(va)
			except:
				pass
		optx.update({ky:va})
	content_type=optx.get('content_type','text/html')
	if content_type:
		hdr="Content-type:{};charset=utf-8\r\n".format(content_type)
		print(hdr)
	return optx

if __name__ == '__main__':
	print(__doc__)



#!/usr/bin/env python3
''' Get series/observations from FRED site
Usage of,
fredSeries.py SERIES [ADDI] [OB_START] [UNITS] 

example,
fredReader.py CPIAUCNS
OR
fredReader.py CPIAUCNS observations 2022-01-02 
OR direct-function-call
python3 -c 'from fredReader import fredReader;df=fredReader("CPIAUCNS",units="pc1",observation_start="2022-01-01");print(df);'

Ref site:
    fred/series - Get an economic data series.
    fred/series/categories - Get the categories for an economic data series.
    fred/series/observations - Get the observations or data values for an economic data series.
    fred/series/release - Get the release for an economic data series.
    fred/series/search - Get economic data series that match keywords.
    fred/series/search/tags - Get the tags for a series search.
    fred/series/search/related_tags - Get the related tags for a series search.
    fred/series/tags - Get the tags for an economic data series.
    fred/series/updates - Get economic data series sorted by when observations were updated on the FRED® server.
    fred/series/vintagedates - Get the latest dates of the series were revised or new data values were released.

'''
import sys
import requests, datetime
import pandas as pd
import numpy as np
import json

def fredReader(series_id,**optx):
	try:
		dropnaTF=optx.pop('dropnaTF',True)
		series_id=series_id.upper()
		d=fredSeries(series_id,**optx)
		df=pd.DataFrame(d["observations"],columns=['date','value'])
		df.set_index(pd.DatetimeIndex(df['date']),inplace=True)
		df.index.name='DATE'
		df['pbdate']=[int(x.strftime("%Y%m%d")) for x in df.index]
		df['series']=series_id
		df['value']=pd.to_numeric(df['value'],errors='coerce')
		df=df[['series','value','pbdate']]
		if dropnaTF:
			df.dropna(inplace=True)
	except Exception as e:
		sys.stderr.write("**ERROR:{}: {}\n".format(series_id,e))
		df = {}
	return df
	
def fredSeries(series_id,addi='observations',file_type='json',api_key=None,debugTF=False,**optx):
	''' Get series from FRED site
		jdata=fredSeries('DGS10')
	    args:
		 series_id:  string as FRED series ID, e.g., "DEXCHUS", "DGS30"
		 addi:  various inquery of data info
		optional args `optx`:
		 units, observation_start
	    return:
		 series observations in json format
	'''
	if not api_key:
		api_key = '{api_key}'.format(**json.load(open('/home/rstudio/.fredapi.json')))
	addi='/'+addi if len(addi) else addi
	#urx='https://api.stlouisfed.org/fred/series{addi}?series_id={series_id}&api_key={api_key}&file_type={file_type}'
	urx='https://api.stlouisfed.org/fred/series{addi}?series_id={series_id}&file_type={file_type}'
	data=dict(api_key=api_key)
	url=urx.format(**locals())
	for k,v in optx.items():
		if not v:
			continue
		url=url+"&{}={}".format(k,v)
	if debugTF:
		sys.stderr.write(url+"\n")
	r=requests.get(url,params=data)
	r.json()
	return r.json()

if __name__ == '__main__':
	args=sys.argv[1:]
	if not args:
		print(__doc__)
		exit(1)
	series= args[0]
	addLst=['categories','observations','release','search','search/tags',
		'search/related_tags','tags','updates','vintagedates']
	addi= 'observations' 
	if len(args)>1 and args[1].lower() in addLst:
		addi=args[1].lower()
	cosd = args[2] if len(args)>2 and len(args[2])==10 else "2022-01-01"
	units = args[3] if len(args)>3 and len(args[3])>2 else ""
	d=fredSeries(series,addi=addi,units=units,observation_start=cosd,debugTF=True)
	if "observations" in d:
		df=pd.DataFrame(d["observations"])
		print(df.tail())
	elif "seriess" in d:
		print(d["seriess"])
	else:
		print(d)



import sys,os
sys.path.append(os.path.dirname(__file__))
from _alan_calc import *
from _alan_date import *
from _alan_str import *
from _alan_pppscf import batch_pppscf,vertex_locator,vertex_mnmx
from yh_chart import *
from yh_hist_batch import *
from yh_predefined import bb_predefined
from find_recent_eps import find_recent_eps
from headline_calc import find_hiloRecord,headline_calc
from headline_sts import headline_hist
from iex_peers import iex_peers,peers_performance,ticker2sectorRanking
from macro_event_yh import run_macro_event_yh
from plot_templates import plot_templates
from record_hilo import get_titlehead
from ticker2label import ticker2label,t2l
from webReader import webReader,streamReader,yhReader
from fredReader import fredReader
import types
unLst=['popen','Call','OLD','TBD','_DEPRECATED','Popen','__']
talan_vars = {k:v for (k,v) in globals().items() if callable(v) and not any(map(k.__contains__,unLst)) and isinstance(v,types.FunctionType) and v.__doc__} 
for k in ['unLst']:
	globals().pop(k,None)



#!/usr/bin/env python3 
# -*- coding: utf-8 -*-
'''
Find leading stocks (up/down > 1%) w.r.t. SP500/DowJone's up/down movement
preparation for generate daily headline context
Usage of
python3 headline_sts.py NDAYS > headline_hist.txt
OR To save result to mongoDB table ara::headline_hist
NDAYS=2; python3 headline_sts.py $NDAYS 1 > headline_hist.txt
OR To get headline_hist for end at specific date:'end'
python -c "from headline_sts import headline_hist as hh;print(hh(ndays=2,saveDB=False,end=20190607).to_csv(index=False))"

also see,
record_hilo.py
Note: python2/3 compatiable
'''
from _alan_calc import sqlQuery,conn2pgdb
from _alan_optparse import parse_opt, subDict
import sys
import ast
if sys.version_info.major == 2:
	from cStringIO import StringIO
else:
	from io import StringIO
import numpy as np
import pandas as pd

# use xqTmp, tkLst
def get_stock_hist(tkLst,ndays=2,xqTmp='',diffTF=False,pgDB=None,end=None,debugTF=False,liveTF=False):
	ticker=tkLst[0]
	xqx = "select pbdate from prc_hist where name='{}' {} order by pbdate DESC limit {}"
	addiClause = '' if end is None else ' AND pbdate<={}'.format(end)
	dtmp = sqlQuery(xqx.format(ticker,addiClause,ndays),pgDB)
	currdb_date = dtmp['pbdate'].iloc[0]
	xdb_date = dtmp['pbdate'].iloc[1]
	if end is not None and int(end)>currdb_date and liveTF is True:
		from yh_hist_batch import yh_hist
		from _alan_date import next_date
		s = next_date(end,dformat='%Y%m%d',days=-5,dtTF=False)
		ranged = "{},{}".format(s,end)
		dtmp = yh_hist(ticker,gap='1d',ranged=ranged,debugTF=debugTF)
		sys.stderr.write("===LIVE dtmp:\n{}\n".format(dtmp.tail()))
		curryh_date = dtmp['pbdate'].iloc[-1]
	else:
		curryh_date = currdb_date
	if curryh_date>currdb_date:
		dx=pd.DataFrame()
		for ticker in tkLst:
			ticker = ticker.replace('.','-')
			dtmp = yh_hist(ticker,gap='1d',ranged=ranged,debugTF=debugTF)[-ndays:]
			dtmp['chgPct'] = dtmp['close'].pct_change().values*100.
			dtmp['chgLevel'] = dtmp['close'].diff(periods=1).values
			dx= pd.concat([dx,dtmp])
		dx.rename(columns={'name':'ticker'},inplace=True)
		dx.reset_index(drop=True,inplace=True)
		sys.stderr.write("===LIVE dx:\n{}\n".format(dx.tail()))
		return dx
	else:
		start = xdb_date
	if len(xqTmp)<1:
		xqTmp = "select * from prc_hist WHERE name in {} and pbdate>={} {} ORDER BY name,pbdate"
	nmLst = [x.replace('.','-') for x in tkLst]
	nmPns="('{}')".format("','".join(nmLst))
	xqr = xqTmp.format(nmPns,start,addiClause)
	dx=sqlQuery(xqr,pgDB)
	dx['ticker']= [x.replace('-','.') for x in dx['name'] ]
	dx['chgPct']= np.empty(dx['ticker'].shape)
	if diffTF is True:
		dx['chgLevel']= np.empty(dx['ticker'].shape)
	for ticker in tkLst:
		dx.loc[dx['ticker']==ticker,'chgPct'] = dx.loc[dx['ticker']==ticker,'close'].pct_change().values*100.
		if diffTF is True:
			dx.loc[dx['ticker']==ticker,'chgLevel'] = dx.loc[dx['ticker']==ticker,'close'].diff(periods=1).values
	return dx

def alloc_yvalue3(vx,thd=[0,0],yvalue=[1,0,-1],openTF=True):
	'''
	Return elements from one of 'yvalue' depending on threshold condition: 'thd'
	'''
	if len(yvalue)<3:
		return vx
	if isinstance(vx,list):
		vx = np.array(vx)
	if hasattr(thd, "__len__"):
		if openTF is True:
			vs = np.where(vx>thd[0],yvalue[0],(np.where(vx<thd[1],yvalue[2],yvalue[1])) )
		else:
			vs = np.where(vx>=thd[0],yvalue[0],(np.where(vx<=thd[1],yvalue[2],yvalue[1])) )
	else:
		if openTF is True:
			vs = np.where(vx>thd,yvalue[0],(np.where(vx<thd,yvalue[2],yvalue[1])) )
		else:
			vs = np.where(vx>=thd,yvalue[0],yvalue[2] )
	return vs

def alloc_yvalue2(vx,thd=0,yvalue=[1,-1],openTF=True):
	'''
	Return elements from one of 'yvalue' depending on threshold condition: 'thd'
	'''
	if len(yvalue)<2:
		return vx
	if isinstance(vx,list):
		vx = np.array(vx)
	if openTF is True:
		vs = np.where(vx>thd,yvalue[0],yvalue[1])
	else:
		vs = np.where(vx>=thd,yvalue[0],yvalue[1])
	return vs

# find top-changes list
def find_top_changes(pbdate=20190405,dx={}):
	cLst = ['ticker','chgPct','pbdate','close']
	if 'chgLevel' in dx.columns:
		cLst = cLst + ['chgLevel']
	df = dx.query('pbdate=={}'.format(pbdate))[cLst]
	df.reset_index(drop=True,inplace=True)
	df['sign'] = np.where(df['chgPct']>0,1,(np.where(df['chgPct']<0,-1,0)) )
	df['UpDn'] = alloc_yvalue3(df['chgPct'],thd=[0.005,-0.005],yvalue=["UP","FLAT","DOWN"])
	caseUp = df.loc[np.where(df['chgPct']>=1)].sort_values(by='chgPct',ascending=False)
	caseDn = df.loc[np.where(df['chgPct']<=-1)].sort_values(by='chgPct',ascending=True)
	return df, caseUp, caseDn

def  get_eqtLst(minMarketCap=50000000000):
	'''
	return equity list for import highlite based on dow_component,sp500_component, yh_quote_curr
	and minimum size (100B) of marketCalp
	'''
	xqTmp = '''SELECT a.ticker FROM sp500_component a, yh_quote_curr b
		WHERE a.ticker=b.ticker AND
		(b."marketCap">{} OR a.ticker in (SELECT ticker FROM dow_component))'''
	xqr = xqTmp.format(minMarketCap)
	try:
		eqtLst =  list(sqlQuery(xqr)['ticker'])
	except Exception as e:
		eqtLst=['AAPL','ABT','ACN','ADBE','AMGN','AMZN','AVGO','AXP','BA','BAC','BRK-B','C','CAT','CMCSA','COST','CRM','CSCO','CVX','DHR','DIS','DOW','FB','GOOG','GOOGL','GS','HD','HON','IBM','INTC','JNJ','JPM','KO','LIN','LLY','LMT','MA','MCD','MDT','MMM','MRK','MSFT','NEE','NFLX','NKE','NVDA','ORCL','PEP','PFE','PG','PM','PYPL','SBUX','T','TMO','TRV','TXN','UNH','UNP','UPS','UTX','V','VZ','WBA','WFC','WMT','XOM']
	return eqtLst

def headline_hist(ndays=2,saveDB=False,end=None,outTF=False,**optx):
	eqtLst = get_eqtLst()
	pgDB=conn2pgdb(dbname='ara')
	# get indices chgPct: idxRtn
	idxLst = ['^GSPC','^DJI','^IXIC','^SOX']
	idxRtnHist = get_stock_hist(idxLst,ndays=ndays,diffTF=True,pgDB=pgDB,end=end)
	if len(idxRtnHist)<1:
		return []

	# get equity chgPct
	eqtRtnHist = get_stock_hist(eqtLst,ndays=ndays,pgDB=pgDB,end=end)
	if len(eqtRtnHist)<1:
		return []

	# get equity chgPct
	pbLst = idxRtnHist.loc[idxRtnHist['ticker']==idxLst[0],'pbdate'].sort_values(ascending=False)[:]
	# redirect stdout
	old_stdout = sys.stdout
	mystdout = StringIO()
	sys.stdout = mystdout
	fwrite=sys.stdout.write

	fwrite("pbdate|Index1|chgPct1|chgLevel1|UpDn1|close1")
	fwrite("|Index2|chgPct2|chgLevel2|UpDn2|close2")
	fwrite("|Index3|chgPct3|chgLevel3|UpDn3|close3")
	fwrite("|Index4|chgPct4|chgLevel4|UpDn4|close4")
	print("|topDate1|topUD1|topLst1|topDate2|topUD2|topLst2|conflict")
	for pbdate in pbLst:
		idxRtn,idxUp,idxDn = find_top_changes(pbdate=pbdate,dx=idxRtnHist)
		eqtRtn,eqtUp,eqtDn = find_top_changes(pbdate=pbdate,dx=eqtRtnHist)
		#print(idxRtn,"\n",eqtUp,"\n",eqtDn)
		if len(idxRtn)<1: 
			sys.stderr.write("skip date: {}\n".format(pbdate))
			continue

		sp500Ret = idxRtn.loc[idxRtn['ticker']=='^GSPC'].iloc[0].to_dict()
		dowRet = idxRtn.loc[idxRtn['ticker']=='^DJI'].iloc[0].to_dict()
		nasdaqRet = idxRtn.loc[idxRtn['ticker']=='^IXIC'].iloc[0].to_dict()
		phlRet = idxRtn.loc[idxRtn['ticker']=='^SOX'].iloc[0].to_dict()
		if np.isnan(sp500Ret['chgPct']):
			sys.stderr.write("skip date: {}\n".format(pbdate))
			continue

		conflict = 0 if dowRet['sign'] == sp500Ret['sign'] else 1

		eqtLx1 = eqtUp if sp500Ret['sign']==1 else eqtDn
		if len(eqtLx1)<1:
			topLst1=[]
			eqtLead1 = {}
		else:
			dtmp=eqtLx1[['ticker','chgPct']].head(3).to_dict(orient='records')
			topLst1=[{x['ticker']:round(x['chgPct'],2)} for x in dtmp]
			#topLst1=["{!r}:{:.2f}".format(*x) for x in (eqtLx1[['ticker','chgPct']].head(3).values)]
			eqtLead1 = eqtLx1.iloc[0].to_dict()
			eqtLead1['topLst1']=topLst1

		eqtLx2 = eqtUp if dowRet['sign']==1 else eqtDn
		if len(eqtLx2)<1:
			topLst2=[]
			eqtLead2 = {}
		else:
			dtmp=eqtLx2[['ticker','chgPct']].head(3).to_dict(orient='records')
			topLst2=[{x['ticker']:round(x['chgPct'],2)} for x in dtmp]
			#topLst2=["{!r}:{:.2f}".format(*x) for x in (eqtLx2[['ticker','chgPct']].head(3).values)]
			eqtLead2 = eqtLx2.iloc[0].to_dict()
			eqtLead2['topLst2']=topLst2

		fwrite("{pbdate}|{ticker}|{chgPct:.2f}|{chgLevel:.0f}|{UpDn}|{close:5g}".format(**sp500Ret))
		fwrite("|{ticker}|{chgPct:.2f}|{chgLevel:.0f}|{UpDn}|{close:5g}".format(**dowRet))
		fwrite("|{ticker}|{chgPct:.2f}|{chgLevel:.0f}|{UpDn}|{close:5g}".format(**nasdaqRet))
		fwrite("|{ticker}|{chgPct:.2f}|{chgLevel:.0f}|{UpDn}|{close:5g}".format(**phlRet))
		if len(eqtLx1)>0:
			fwrite("|{pbdate}|{UpDn}|{topLst1}".format(**eqtLead1))
		if len(eqtLx2)>0:
			fwrite("|{pbdate}|{UpDn}|{topLst2}".format(**eqtLead2))
		fwrite("|{conflict}\n".format(conflict=conflict))

	# redirect the stdout to string and convert it to dataframe
	xstr = mystdout.getvalue()
	sys.stdout = old_stdout
	fwrite=sys.stdout.write
	if outTF:
		fwrite(xstr)
	df = pd.read_csv(StringIO(xstr),sep='|')
	from _alan_calc import save2pgdb
	from _alan_str import write2mdb
	if saveDB==True:
		tablename="headline_hist"
		sys.stderr.write("Save to {}\n".format(tablename))
		#save2pgdb(df,db=pgDB,tablename=tablename)
		clientM=None
		zpk={"ticker","pbdate"}
		mobj,clientM,msg = write2mdb(df,clientM,tablename=tablename,zpk=zpk)

	if 'topDict' in optx and optx['topDict']==True:
		if 'topLst1' in df:
			df['topLst1']=[ ast.literal_eval(x) if hasattr(df['topLst1'], "__len__") else {} for x in df['topLst1'] ]
		if 'topLst2' in df:
			df['topLst2']=[ ast.literal_eval(x) if hasattr(df['topLst2'], "__len__") else {} for x in df['topLst2'] ]

	df['chg1']=df['chgLevel1']
	df['chg2']=df['chgLevel2']
	df['allUpDn']=0
	for j in range(df.shape[0]):
		if all([ x>0 for x in df[['chgPct1','chgPct2','chgPct3']].iloc[j] ]) :
			allUpDn = 1
		elif all([ x<0 for x in df[['chgPct1','chgPct2','chgPct3']].iloc[j] ]) :
			allUpDn = -1
		else:
			allUpDn = 0
		df.loc[df.index[j],'allUpDn']=allUpDn
	return df

if __name__ == '__main__':
	description="""Find leading stocks (up/down > 1%) w.r.t. SP500/DowJone's up/down movement"""
	(opts, args)=parse_opt(sys.argv,ndays=2,outTF=True,saveDB=False)
	df = headline_hist(topDict=True,**opts)
	print(df.to_dict(orient='records'))



#!/usr/bin/env python
""" 
pull yahoo calendar by event [economic|earnings|splits|ipo]
and save info to table: [earnings_yh], [economic_yh]...
OR pull yahoo earnings history for specific stock 'ticker'
and save info to table: [earnings_hist_yh]

Usage of:
python macro_event_yh.py --type=earnings --day=20190801
OR
python macro_event_yh.py --ticker=aapl
OR
python macro_event_yh.py --type=ipo --day=20190612
OR
python macro_event_yh.py --type=splits --day=20190813
OR
printf "aapl\nibm"| python macro_event_yh.py --type=earnings --ticker=-

Ref_uri:
https://finance.yahoo.com/calendar/earnings?day=2019-01-24
OR
https://finance.yahoo.com/calendar/earnings?symbol=AAPL
OR for specific date
https://finance.yahoo.com/calendar/{event}?day={yyyy-mm-dd}
OR for specific ticker
https://finance.yahoo.com/calendar/earnings?symbol={ticker}

Also see: eps_nasdaq.py
Last mod., Tue Jan 29 14:08:11 EST 2019
"""
import sys
from optparse import OptionParser
import pandas as pd
from sqlalchemy import create_engine, MetaData
from datetime import datetime
import numpy as np
from _alan_str import upsert_mdb
from io import StringIO
import requests
headers={'Content-Type': 'text/html', 'Accept': 'application/json', 'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'}
headers={'Content-Type': 'text/html', 'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'}

def get_datestr(xdate):
	if xdate is None:
		return datetime.today().strftime('%Y-%m-%d')
	if isinstance(xdate,float):
		xdate = str(int(xdate))
	elif isinstance(xdate,(int,np.integer)):
		xdate = str(xdate)
	if xdate.isdigit() and len(xdate)==8:
		d = xdate
		xdate = '{}-{}-{}'.format(d[:4],d[4:6],d[-2:])
	elif '/' in xdate:
		xdate.replace('/','-')
	return xdate

def earnings_morningstar(xdate=None,xtype='earnings',uri=None,dbname='ara',tablename='earnings_morningstar',debugTF=False,**optx):
	""" Usage of:
	python -c "from macro_event_yh import earnings_morningstar as em;em(xdate=20190430)"
	"""
	xdate=get_datestr(xdate)
	#http://www.morningstar.com/earnings/Handler/GetEarningsCalendar.ashx?rptDate=2019-05-01&r=68265
	if uri is None:
		urx="http://www.morningstar.com/earnings/Handler/GetEarningsCalendar.ashx?rptDate={}"
		uri = urx.format(xdate)
	ret=pd.read_html(uri)
	df = ret[1]
	v =[str(x.split()[-1].encode('ascii','ignore')) for x in df[u'Company Name']]
	df['ticker'] = v
	df['pbdate'] = int(xdate.replace('-',''))
	zpk = {'ticker','pbdate'} if xtype == 'earnings' else {'event','pbdate'}
	clientM=None
	print( v)
	mobj,clientM,err_msg = upsert_mdb(df,clientM,dbname=dbname,tablename=tablename,zpk=zpk)
	return df

def get_calendar_info(xdate,xtype='economic',uri=None,classname=None,debugTF=False,**optx):
	xdate = get_datestr(xdate)
	if uri==None:
		uri="https://finance.yahoo.com/calendar/{}?day={}".format(xtype,xdate)
	if classname==None:
		classname="data-table W(100%) Bdcl(c) Pos(r) BdB Bdc($c-fuji-grey-c)"
	url=uri
	for j in [0,1,2,3,4,5]:
		try:
			if j>0:
				url = uri + '&offset={}00&size=100'.format(j)
			sys.stderr.write( "{}\n.....\n".format(url)+"\n" )
			ret = requests.Session().get(url,headers=headers,verify=False)
			xa = pd.read_html(StringIO(ret.text),index_col=0,header=0)
			#xa=pd.read_html(url,index_col=0,header=0)
			if len(xa[0])<1:
				break
			if j>0:
				df = pd.concat([df,xa[0]])
			else:
				df=xa[0]
		except Exception as e:
			sys.stderr.write( '**ERROR: {}\n{}\n'.format(str(e),url)+"\n" )
			break
	df['pbdate'] = int(xdate.replace('-',''))
	if debugTF:
		sys.stderr.write( 'DF:\n{}'.format(df)+"\n" )
	sys.stderr.write( 'Total list {}'.format(len(df))+"\n" )
	return df

def search_splits(xdate,xtype='ipo',uri=None,classname=None,debugTF=False,**optx):
	df = get_calendar_info(xdate,xtype=xtype,uri=uri,classname=classname)
	if len(df)<1:
		return {}
	df.columns = [x.replace(' ','').replace('?','') for x in df.columns]
	df = df.reset_index()
	return df

def search_ipo(xdate,xtype='ipo',uri=None,classname=None,debugTF=False,**optx):
	df = get_calendar_info(xdate,xtype=xtype,uri=uri,classname=classname)
	if len(df)<1:
		return {}
	df = df.query("Currency=='USD' & Actions=='Priced'")
	if len(df)<1:
		return {}
	df.columns = [x.replace(' ','') for x in df.columns]
	#hdr=['Symbol','Company','Exchange','Date','Price Range','Price','Currency','Shares','Actions']
	#df.columns = hdr
	df = df.reset_index()
	return df

def search_economic(xdate,xtype='economic',uri=None,classname=None,debugTF=False,**optx):
	df = get_calendar_info(xdate,xtype=xtype,uri=uri,classname=classname)
	if len(df)<1:
		return {}
	df = df.query("Country=='US' & Actual!='-'")
	hdr=['event','Country','EventTime','ForPeriod','Actual','MarketExpectation','Previous','RevisedFrom','pbdate']
	df = df.reset_index()
	df.columns = hdr
	return df

def search_earnings(xdate,xtype='earnings',uri=None,classname=None,debugTF=False,**optx):
	df = get_calendar_info(xdate,xtype=xtype,uri=uri,classname=classname)
	if len(df)<1:
		return {}
	hdr=['ticker','Company', 'CallTime', 'estimatedEPS', 'actualEPS', 'EPSSurprisePCT','pbdate']
	df = df.reset_index()
	df.columns = hdr
	return df

def search_earnings_by_symbol(ticker,uri=None,xtype='earnings',debugTF=False,**optx):
	url=uri
	if url is None:
		urx='https://finance.yahoo.com/calendar/earnings?symbol={}'
		url = urx.format(ticker)
	sys.stderr.write( "Pulling\n{}\n...".format(url)+"\n" )
	try:
		ret = requests.Session().get(url,headers=headers,verify=False)
		xf = pd.read_html(StringIO(ret.text))
		#xf = pd.read_html(uri,index_col=False,header=0)
	except Exception as e:
		sys.stderr.write( '**ERROR: {}:{}  @{}()'.format(ticker,str(e),'search_earnings_by_symbol')+"\n" )
		return {};
	if len(xf)<1:
		return {};
	df = xf[0]
	if debugTF:
		sys.stderr.write('{}'.format(df)+"\n" )
		sys.stderr.write('{}'.format(df.columns)+"\n" )
	if isinstance(df['EPS Estimate'].iloc[0],str):
		df = df.drop(df[df['EPS Estimate']=='-'].index).reset_index(drop=True)
		df['estimatedEPS'] = [ float(x) if x!='-' else None for x in df['EPS Estimate'].values ]
	if 'Reported EPS' in df.columns and isinstance(df['Reported EPS'].iloc[0],str):
		df['actualEPS'] = [ float(x) if x!='-' else None for x in df['Reported EPS'].values ]
	df.rename(columns={'Symbol':'ticker'},inplace=True)
	if 'Earnings Date' in df.columns:
		edt = [ datetime.strptime(x,'%b %d, %Y, %I %p%Z') for x in df['Earnings Date'].values ]
		df['pbdate']  = [ int(x.strftime('%Y%m%d')) for x in edt ]
		df.loc[:,'Earnings Date']  = edt
	return df

def run_macro_event_yh(uri=None,xdate=None,saveDB=True,tablename=None,wmode='replace',xtype='earnings',clientM=None,hostname='localhost',dbname=None,ticker=None,debugTF=False,**optx):
	""" find macro event from yahoo
	"""
	pgDB = None
	if ticker is not None:
		urx='https://finance.yahoo.com/calendar/earnings?symbol={}'
		uri = urx.format(ticker)
		funcName = 'search_earnings_by_symbol'
		tablename = "{}_hist_yh".format(xtype)
	else:
		funcName = "search_{}".format(xtype)
		tablename = "{}_yh".format(xtype) if not tablename else tablename
	if funcName in globals():
		searchFunc = globals()[funcName]
		dfyc=searchFunc(xdate,uri=uri,xtype=xtype)
		if 'Symbol' in dfyc and 'ticker' not in dfyc:
			dfyc['ticker']=dfyc['Symbol']
		if len(dfyc)<1:
			return {}
	else:
		return {}
	
	if any([saveDB is False]):
		sys.stderr.write('{}'.format(dfyc.to_csv(index=False,sep='|'))+"\n" )
	else:
		zpk = {'ticker','pbdate'} if xtype in ['earnings','ipo','splits'] else {'event','pbdate'}
		mobj,clientM,err_msg = upsert_mdb(dfyc,clientM,dbname=dbname,tablename=tablename,zpk=zpk)
		sys.stderr.write( "{}\n...\n{}\n saved to {}:{}".format(dfyc.head(1),dfyc.tail(1),clientM,tablename)+"\n" )
	if pgDB is not None:
		dfyc.to_sql(tablename,pgDB,index=False,schema='public',if_exists=wmode)
	return dfyc

def opt_macro_event_yh(argv,retParser=False):
	parser = OptionParser(usage="usage: %prog [option]", version="%prog 0.1",
		description="get up-to-date calendar events")
	parser.add_option("","--uri",action="store",dest="uri",
		help="uri (default: None)")
	parser.add_option("","--ticker",action="store",dest="ticker",
		help="ticker (default: None), for [earnings] type ONLY")
	parser.add_option("","--day",action="store",dest="xdate",
		help="YYYYMMDD (default: today)")
	parser.add_option("","--type",action="store",dest="xtype",default="earnings",
		help="calendar type [economic|earnings](default: earnings)")
	parser.add_option("-d","--database",action="store",dest="dbname",default="ara",
		help="database (default: ara)")
	parser.add_option("","--host",action="store",dest="hostname",default="localhost",
		help="db host (default: localhost)")
	parser.add_option("-t","--table",action="store",dest="tablename",
		help="db tablename")
	parser.add_option("-w","--wmode",action="store",dest="wmode",default="replace",
		help="db table write-mode [replace|append|fail] (default: replace)")
	parser.add_option("","--no_database_save",action="store_false",dest="saveDB",default=True,
		help="no save to database (default: save to database)")
	parser.add_option("","--debug",action="store_true",dest="debugTF",default=False,
		help="DEBUG flag (default: False)")
	(options, args) = parser.parse_args(argv[1:])
	if retParser is True:
		return parser
	return (vars(options), args)

if __name__ == '__main__':
	opt,args =opt_macro_event_yh(sys.argv)
	if opt['ticker'] is not None:
		tkLst=opt['ticker'].split(',')
		if tkLst[0]=='-':
			tkLst = sys.stdin.read().strip().upper().split()
		for ticker in tkLst:
			opt['ticker'] = ticker
			opt['xtype'] = 'earnings'
			try:
				run_macro_event_yh(**opt)
			except Exception as e:
				sys.stderr.write( '**ERROR: {} @ {}'.format(ticker,str(e))+"\n" )
	else:
		run_macro_event_yh(**opt)



#!/usr/bin/env python3
""" Program: prc_temp_DN.py
    Description: Script to download Yahoo history
    Usage of:
	# from a list of SYMBOLS
		python prc_temp_DN.py SYMBOL1 SYMBOL2 ...
	OR
	# from stdin
		printf "SELECT ticker FROM mapping_ticker_cik where act_code=1"| psql.sh -d ara -At | python prc_temp_DN.py 
	to get [SYMBOL1,SYMBOL2,...] history
	and save them into [dbname]:prc_temp_yh
	then run prc_yh2hist.sql to insert additional to [prc_hist] (optional)
	OR
		printf "SELECT series FROM mapping_series_label"| psql.sh -d ara -At | python prc_temp_DN.py --table=macro_temp_fred
	then run fred2hist.sql to insert additional to [macro_hist_fred] (optional)
	OR
		cat currency.list | python prc_temp_DN.py -d ara --no_database_save
	to view history
    Note:
	To consolidate version for prc_temp_cn_DN.py & prc_temp_cn_DN.py.deprecated.
	This program will retire the following programs:
	prc_temp_av_DNLD.py
	prc_temp_batch.py
	prc_temp_cn.tradego.py
	prc_temp_DNLD.py
	prc_temp_yh_DNLD.py
	prc_temp_yh.py
    add '-' for SYMBOL1 as input from stdin
    direct use _alan_calc.data_from_web function 
    Last mod., Tue Apr 16 09:18:04 EDT 2019
    Last mod., Mon Oct 11 13:05:32 EDT 2021
"""
import sys
from optparse import OptionParser
import datetime
import pandas as pd
from sqlalchemy import create_engine
from _alan_calc import data_from_web

def batch_prc_temp(tkLst,dbname='ara',output=None,tablename='prc_temp_yh',
	src='yh',days=1,end=None,hostname='localhost',wmode='replace',
	tsTF=True,sep='|',indexTF=False,debugTF=False,saveDB=False,start=None):
	""" Get yahoo price history via [pandas_datareader.data] 
		from a list of tickers:[tkLst]
		and save them into table:[tablename] and database URL: [dbURL]
	"""
	#for (ky,va) in opts.items():
	#	exec("{}=va".format(ky))
	try:
		dbURL='postgresql://sfdbo@{}:5432/{}'.format(hostname,dbname)
		engine = create_engine(dbURL) 
	except:
		sys.stderr.write( "***DB ERROR:{}".format(sys.exc_info()[1])+'\n')
		engine = None
	btime = datetime.datetime.now()
	sys.stderr.write("BEGIN----- @ {} -----BEGIN".format(btime)+'\n')
	rmode= wmode
	for i,ticker in enumerate(tkLst):
		sys.stderr.write("=== pulling {}: {}".format(i,ticker)+'\n')
		try:
			symbol=ticker.upper()
			df=data_from_web(symbol,start=start,end=end,days=days,src=src,debugTF=debugTF)
			if df is None or len(df)<1:
				continue
			if(engine is not None and saveDB is True) :
				df.to_sql(tablename,engine,schema='public',index=False,if_exists=rmode)
				rmode= "append"
			if debugTF is True:
				sys.stderr.write("{}".format( df.tail(2))+'\n')
			if output is not None and len(df)>0:
				sys.stdout.write("{}".format(df.to_csv(sep=sep))) 
			#yield df
		except Exception as e:
			sys.stderr.write( "***ERROR {}:{} @ prc_temp_DN.batch_prc_temp():{}".format(i,symbol,str(e))+'\n')
			continue

	etime = datetime.datetime.now()
	sys.stderr.write("Total time: {}".format(etime-btime)+'\n')
	sys.stderr.write("END----- @ {} -----END".format(etime)+'\n')
	if(engine is not None) :
		engine.dispose()

def opt_prc_temp_DN(argv,retParser=False):
	""" command-line options initial setup
	    Arguments:
		argv:   list arguments, usually passed from sys.argv
		retParser:      OptionParser class return flag, default to False
	    Return: (options, args) tuple if retParser is False else OptionParser class
	"""
	parser = OptionParser(usage="usage: %prog [option] SYMBOL1 ...", version="%prog 0.65",
		description="Script to download Yahoo/FRED history")
	parser.add_option("","--days",action="store",dest="days",default=1,type=int,
		help="number of days from endDate (default: 1)")
	parser.add_option("-s","--start",action="store",dest="start",
		help="start YYYY-MM-DD (default: 1-day-ago)")
	parser.add_option("-e","--end",action="store",dest="end",
		help="end YYYY-MM-DD (default: today)")
	parser.add_option("","--src",action="store",dest="src",default="yh",
		help="source [fred|iex|yahoo](default: yh)")
	parser.add_option("-d","--database",action="store",dest="dbname",default="ara",
		help="database (default: ara)")
	parser.add_option("","--host",action="store",dest="hostname",default="localhost",
		help="db host (default: localhost)")
	parser.add_option("-t","--table",action="store",dest="tablename",default="prc_temp_yh",
		help="db tablename (default: prc_temp_yh)")
	parser.add_option("-w","--wmode",action="store",dest="wmode",default="replace",
		help="db table write-mode [replace|append] (default: replace)")
	parser.add_option("","--no_database_save",action="store_false",dest="saveDB",default=True,
		help="no save to database (default: save to database)")
	parser.add_option("-o","--output",action="store",dest="output",
		help="OUTPUT type [csv|html|json] (default: no output)")
	parser.add_option("","--no_datetimeindex",action="store_false",dest="tsTF",default=True,
		help="no datetime index (default: use datetime)")
	parser.add_option("","--show_index",action="store_true",dest="indexTF",default=False,
		help="show index (default: False) Note, OUTPUT ONLY")
	parser.add_option("","--sep",action="store",dest="sep",default="|",
		help="output field separator (default: |) Note, OUTPUT ONLY")
	parser.add_option("","--debug",action="store_true",dest="debugTF",default=False,
		help="debugging (default: False)")
	(options, args) = parser.parse_args(argv[1:])
	if retParser is True:
		return parser
	return (vars(options), args)

if __name__ == '__main__':
	(options, args)=opt_prc_temp_DN(sys.argv)
	if len(args)==0 or (len(args)==1 and args[0]=='-'):
		sys.stderr.write("\nRead from pipe\n"+'\n')
		tkLst = sys.stdin.read().strip().split()
	elif len(args)==1 and ',' in args[0]:
		tkLst = args[0].strip().split(',')
	else:
		tkLst = args
		options['saveDB']=False
	r = batch_prc_temp(tkLst,**options)



#!/usr/bin/env python3
'''
pull most recent ESP 
table required: MDB::earnings_hist_yh,earnings_zacks 
for last season or latest one respectively
Usage of:
python -c "ticker='AAPL';from find_recent_eps import find_recent_eps as findEps;df=findEps(ticker);print(df)"
'''
from _alan_str import find_mdb, num_en2cn
from _alan_date import next_date,s2dt
from _alan_calc import sqlQuery
import datetime
import pandas as pd
import numpy as np
import sys
import math

def select_eps_hist(ticker='AAPL',pbdate=None,types='earnings',dbname='ara',**optx):
	if types.lower() not in ['earnings','eps']:
		return {}
	debugTF=optx.get('debugTF',False)
	tablename=optx.get('tablename','earnings_zacks')
	pbdate = int(next_date(pbdate,dformat='%Y%m%d',dtTF=False,days=-2))
	jobj = {"pbdate":{'$lt':pbdate},"ticker":ticker}
	ret,_,errmsg = find_mdb(jobj,dbname=dbname,tablename=tablename,sortLst=['pbdate'],ascendingTF=False)
	df = pd.DataFrame(ret)
	df = df.rename(columns=dict(Reported="actualEPS",Estimate="estimatedEPS"))
	df = df.dropna(subset=['actualEPS'])
	if debugTF:
		sys.stderr.write("EPS {} of {}:\n".format(ticker,tablename))
		sys.stderr.write("{}\n".format(df))
	return df

def select_eps_hist_yh(ticker='AAPL',pbdate=None,types='earnings',dbname='ara',**optx):
	debugTF=optx.get('debugTF',False)
	if types.lower() in ['earnings','eps']:
		tablename='earnings_zacks'
	elif types.lower() == 'stats':
		tablename='qS_keyStatistics'
		jobj = {"ticker":ticker}
		ret,_,errmsg = find_mdb(jobj,dbname=dbname,tablename=tablename)
		if len(ret)<1:
			return {}
		df = pd.DataFrame(ret)
		return df
	else:
		return {}
	pbdate = int(next_date(pbdate,dformat='%Y%m%d',dtTF=False,days=-2))
	jobj = {"pbdate":{'$lt':pbdate},"ticker":ticker}
	ret,_,errmsg = find_mdb(jobj,dbname=dbname,tablename=tablename,sortLst=['pbdate'],ascendingTF=False)
	df = pd.DataFrame(ret)
	if debugTF:
		sys.stderr.write("EPS {} of {}:\n".format(ticker,tablename))
		sys.stderr.write("{}\n".format(df))
	if 'actualEPS' not in df:
		return df
	df = df.dropna(subset=['actualEPS'])
	df = df[~df['actualEPS'].isin(['-'])]
	df['actualEPS']=df['actualEPS'].astype(float)
	return df

# DEPRECATED, USE yahoo number
def select_eps_histIEX(ticker='UNP',types='eps',nrow=1):
	""" select eps/revenue/profitMaring from PSQL tables
	"""
	if types=='eps':
		xqTmp="SELECT * from iex_earnings_hist where ticker='{}' order by pbdate DESC limit {}"
	elif types=='stats':
		xqTmp="SELECT * from iex_stats_hist where ticker='{}' order by pbdate DESC limit {}"
	else:
		xqTmp="SELECT * from iex_financials_hist where ticker='{}' and freq='Q' order by pbdate DESC limit {}"
	xqr=xqTmp.format(ticker,nrow)
	df = sqlQuery(xqr)
	if len(df)<1:
		return {}
	return df

def calc_pchg(*eps):
	""" 
	calc pctChange if denominator eps[1] is positive
	and return (Change Percent, True
	otherwise (change differrene, False)
	"""
	if eps[1]<=0:
		pchg=eps[0]*1.0-eps[1]
	else:
		pchg=(eps[0]*1.0/eps[1]-1)
	return (pchg, eps[1]>0)

def find_past_eps(ticker='UNP',cdt=None,debugTF=False):
	dd=dict(ticker=ticker)
	dh = select_eps_hist(ticker=ticker,pbdate=cdt,debugTF=debugTF)
	if len(dh)>1:
		if len(dh)>4:
			xeps = dh['actualEPS'].iloc[4]
			dd.update(eps_chgFreq='Y')
		else:
			xeps = dh['actualEPS'].iloc[1]
			dd.update(eps_chgFreq='Q')
		ceps = dh['actualEPS'].iloc[0]
		pchg,chgTF = calc_pchg(ceps,xeps)
		dd.update(dh.iloc[0].to_dict())
		dd.update(eps_pctChange=pchg,eps=ceps,prev_eps=xeps,eps_isPct=chgTF)
		dd.update(latest2weekTF=True)
		dd.update(past_eps = dh[['actualEPS', 'pbdate']].set_index('pbdate'))
		#sys.stderr.write("dh:\n{}\n".format(dh[['actualEPS', 'pbdate']]))
	dh = select_eps_hist(ticker=ticker,types='financials',nrow=2,debugTF=debugTF)
	if len(dh)>1:
		xeps = dh['totalRevenue'].iloc[1]
		ceps = dh['totalRevenue'].iloc[0]
		pchg,chgTF = calc_pchg(ceps,xeps)
		dd.update(revenue_pctChange=pchg,revenue=ceps,prev_revenue=xeps)
		ceps,xeps = (dh['operatingIncome']/dh['operatingRevenue']).values[:2]
		#ceps,xeps = (dh['profitMargin']).values[:2]
		if not math.isnan(ceps):
			pchg,chgTF = calc_pchg(ceps,xeps)
			dd.update(profitMargin_pctChange=pchg,profitMargin=ceps,prev_profitMargin=xeps,profitMargin_isPct=chgTF)
	return(dd)

def find_latest_eps(ticker='UNP',cdt=None,dbname='ara',debugTF=False):
	"""
	Find eps since monday of the last week
	"""
	if not cdt:
		cdt = next_date()
	elif not isinstance(cdt,datetime.date):
		cdt = next_date(cdt)
	gap = -1 if cdt.weekday()==0 else -2
	xdt = next_date(cdt,weeks=gap,weekday=0) # monday of the last week
	sundt = next_date(xdt,days=6) # coming sunday from xdt(monday)
	xdate=int(xdt.strftime('%Y%m%d'))
	findDct={"pbdate":{'$gte':xdate},"ticker":ticker}
	# use zacks for up-to-date EPS
	ret,_,errmsg = find_mdb(jobj=findDct,dbname=dbname,tablename='earnings_zacks',field=['ticker','Estimate','Reported','pbdate','Time'],sortLst=['pbdate'],ascendingTF=False)
	if len(ret)>0:
		dd = ret[0]
		dd['eps'] = float(dd['Reported'])
		pdt = s2dt(dd['pbdate'])
		if pdt<=xdt: # discard any data less Monday of the week 
			sys.stderr.write("Stale news:{} < {}, disregard!\n".format(pdt,xdt))
			return {}
		currWeekTF= (pdt>sundt)
		todayTF= (pdt.day==cdt.day)
		dd.update(weekday=pdt.weekday()+1,currWeekTF=currWeekTF,todayTF=todayTF)
		pbdate=dd['pbdate']
	else:
		return {}
	dh = select_eps_hist(ticker=ticker,pbdate=pbdate,debugTF=debugTF)
	if len(dh)>0:
		ceps = dd['eps']
		if dh.shape[0]>3:
			xeps = dh['actualEPS'].iloc[3] 
			dd.update(eps_chgFreq='Y')
		else:
			xeps = dh['actualEPS'].iloc[0] 
			dd.update(eps_Freq='Q')
		pchg,chgTF = calc_pchg(ceps,xeps)
		dd.update(eps_pctChange=pchg,prev_eps=xeps,eps_isPct=chgTF)
		dd.update(latest2weekTF=True)
		d1 = dh[['actualEPS', 'pbdate']].set_index('pbdate')
		d1.loc[pbdate] = {'actualEPS':ceps}
		dd.update(past_eps = d1.sort_index())
	dh = select_eps_hist(ticker=ticker,types='financials',debugTF=debugTF)
	if len(dh) and 'revenue' in dd:
		revenueStr = dd['revenue']
		revenue = num_en2cn( revenueStr,numTF=True)
		pchg = revenue/dh['totalRevenue'].iloc[0]-1
		dd.update(revenue_pctChange=pchg,prev_revenue=xeps)
		dd.update(revenue=revenue,revenueStr=revenueStr)
	return(dd)

def find_recent_eps(ticker='UNP',cdt=None,debugTF=False):
	df = find_latest_eps(ticker,cdt=cdt,debugTF=debugTF)
	if len(df)<1:
		sys.stderr.write("{}\n".format("Latest EPS not found!"))
		df = find_past_eps(ticker,cdt=cdt,debugTF=debugTF)
	if 'pbdate' in df:
		pbdate=int(df['pbdate'])
		df.update(pbdate=pbdate)
	return df

if __name__ == '__main__':
	ticker = sys.argv[1] if len(sys.argv)>1 else 'AMZN'
	cdt = sys.argv[2] if len(sys.argv)>2 else None
	dd =find_recent_eps(ticker,cdt=cdt,debugTF=True)
	#sys.stdout.write("{}\n".format(dd))
	# serializing datetime and bool and dataframe for json
	import json,datetime
	#dtCvt = lambda x: x.__str__() if isinstance(x, (bool,datetime.datetime)) else ''
	def dtCvt(obj): 
		if isinstance(obj, (bool,datetime.datetime)):
			o = obj.__str__()
		elif isinstance(obj, pd.DataFrame):
			o= obj.to_json(orient='records')
		else:
			o= ''
		return o
	sys.stdout.write("{}\n".format(json.dumps(dd,default=dtCvt)))




#!/usr/bin/env python3
# get peer group based on ticker
# Usage of:
# python -c "from iex_peers import ticker2sectorRanking as t2sr;print t2sr(ticker='AMZN',sector='Consumer Discretionary')"
# python -c "from iex_peers import peers_performance as ppf;print ppf(tkLst=['AMZN','BABA','JD','EBAY','MELI','ORLY','XLY'])"

import sys
import numpy as np
import pandas as pd
from _alan_calc import sqlQuery

def df2out(df,output='',indexTF=False,sep='|',orient='records',formatters=None,float_format=None,justify=None,border=None,table_id=None,columns=None,force_ascii=False):
	""" datafame to alternative output format
	"""
	if output=='csv':
		return df.to_csv(index=indexTF,sep=sep)
	elif output=='json': # orient: [split|records|index|columns|values]
		return df.to_json(orient=orient,force_ascii=force_ascii).encode('utf8')
	elif output=='dict': # orient: [dict|list|series|split|records|index]
		return df.to_dict(orient=orient)
	elif output=='html':
		return df.to_html(index=indexTF,formatters=formatters,
			float_format=float_format,justify=justify,border=border,
			table_id=table_id,columns=columns)
	else:
		return df

def peers_performance(tkLst=[],xqTmp='',debugTF=True):
	'''
	return peRatio's for comparison based on tickers: 'tkLst'
	'''
	if not tkLst:
		return []
	if not xqTmp:
		xqTmp="""select ticker,"changePercent"/100.0 as pchg, "trailingPE" as "peRatio" from yh_quote_curr where ticker in ('{}') ORDER BY pchg DESC"""
	xqr = xqTmp.format("','".join(tkLst) )
	ret = sqlQuery(xqr)
	#ret = ret.dropna()
	return ret

def merge_ticker_cik(tkLst=[],colStr='ticker,company_cn',dbname='ara',tablename='mapping_ticker_cik'):
	xqTmp="""Select {colStr} from "{tablename}" where ticker in ('{tkQStr}')"""
	tkQStr="','".join(tkLst)
	xqr = xqTmp.format(**locals())
	df = sqlQuery(xqr,dbname=dbname)
	return df

def ticker2sectorRanking(ticker,sector='Information Technology',xqTmp='',debugTF=True):
	'''
	return ranking of 'ticker' in the 'sector' 
	'''
	if not xqTmp:
		xqTmp="""Select ticker,"marketCap" from yh_quote_curr where  "marketCap">=(SELECT "marketCap" FROM yh_quote_curr WHERE ticker='{ticker}') and ticker in (SELECT ticker from mapping_ticker_cik where sector='{sector}') ORDER BY "marketCap" DESC"""
	xqr = xqTmp.format(ticker=ticker,sector=sector)
	ret = sqlQuery(xqr)
	return ret

def ticker_peers(ticker,xqTmp='',debugTF=True,j=0):
	'''
	USE
	SELECT a.*,b."marketCap" as marketCap, b."trailingPE" as peRatio FROM (select ticker,sector,industry from "yh_summaryProfile" where industry in (select industry from "yh_summaryProfile" where ticker='{}')) as a, yh_quote_curr b WHERE a.ticker=b.ticker  ORDER BY marketCap DESC
	'''
	if len(xqTmp)<1:
		xqLst=(
		"""SELECT a.*,b."marketCap" as marketCap, b."trailingPE" as peRatio FROM (select ticker,sector,industry from "yh_summaryProfile" where industry in (select industry from "yh_summaryProfile" where ticker='{}')) as a, yh_quote_curr b WHERE a.ticker=b.ticker  ORDER BY marketCap DESC""",
		"""SELECT a.*,b."marketCap" as marketCap, b."trailingPE" as peRatio FROM (select ticker,sector,industry from "yh_summaryProfile" where sector in (select sector from "yh_summaryProfile" where ticker='{}')) as a, yh_quote_curr b WHERE a.ticker=b.ticker  ORDER BY marketCap DESC""",
		"""SELECT a.*,b.\"marketCap\" as marketCap, b.\"trailingPE\" as peRatio FROM (SELECT etfname as ticker,sector_alias as sector,NULL as industry FROM spdr_sector WHERE sector_alias in (select sector from "yh_summaryProfile" where ticker='{}' )) as a, yh_quote_curr b WHERE a.ticker=b.ticker  ORDER BY marketCap DESC""")
	xqTmp = xqLst[j]
	xqr=xqTmp.format(ticker)
	if debugTF is True:
		sys.stderr.write(xqr+"\n")
	ret = sqlQuery(xqr)
	if len(ret)<5 and j<len(xqLst)-1:
		ret=ticker_peers(ticker,debugTF=debugTF,j=j+1)
	return ret

def get_peers(ticker,xqTmp='',top=6,lb=-2,nb=5,debugTF=True):
	try:
		ret = ticker_peers(ticker,debugTF=debugTF)
	except Exception as e:
		sys.stderr.write("**ERROR:{}\n".format(str(e)))
		return {}
	if len(ret)<1:
		return {}
	idx = ret.query('ticker=={!r}'.format(ticker)).index[0]
	a = np.arange(top)
	b = max(0,idx+lb)+np.arange(nb)
	c = np.unique(np.concatenate((a, b)))
	d = np.intersect1d(ret.index,c)
	dx1 = ret.loc[d,:]
	# add sector etf peer
	dx2 = ticker_peers(ticker,debugTF=debugTF,j=2) # apply sector directly
	df = pd.concat([dx1,dx2],ignore_index=False)
	df['idx']=0
	df.reset_index(drop=True,inplace=True)
	idx = df.query('ticker=={!r}'.format(ticker)).index[0]
	df.loc[idx,'idx']=1
	df.rename(columns={'ticker':'peers'},inplace=True)
	df['ticker'] = ticker
	if debugTF is True:
		sys.stderr.write("{}".format((a,b,c,d)))
		sys.stderr.write("{}".format(df))
	return df

def yh_peers(tkLst=[],debugTF=False,peersLstTF=False):
	""" Get peer group based on tickers: 'tkLst'
	"""
	if not isinstance(tkLst,list):
		tkLst=[tkLst]
	if len(tkLst)<1:
		sys.stderr.write("tkLst:{} is empty.\n".format(tkLst))
		return {} if peersLstTF else ({},{},{})
	df=pd.DataFrame()
	peerLst=[]
	for tkX in tkLst:
		dx = get_peers(tkX,debugTF=debugTF)
		if len(dx)<1:
			continue
		xp = { 'peers' :[str(x) for x in dx['peers'].values if x!=tkX] }
		xp.update(ticker=tkX)
		peerLst.append(xp)
		df = pd.concat([df,dx],ignore_index=False)
	if 'peers' not in df:
		ret= [] if peersLstTF else df,[],{}
		return ret
	tqLst=df['peers'].values
	dq = merge_ticker_cik(tqLst)
	if len(dq)>0:
		dq.rename(columns={'ticker':'peers'},inplace=True)
		df = df.merge(dq,on='peers',how='left')
	if peersLstTF:
		return peerLst
	if len(df)>0: 
		peerInfo=dict(peersList=','.join(df['peers'].values),peers=list(df['peers']))
		peerInfo.update(peers_cn=list(df['company_cn']))
	else :
		peerInfo=dict(peersList='',peers=[],peers_cn=[])
	return df,peerLst,peerInfo

def iex_peers(tkLst=[],debugTF=False,peersLstTF=False):
	""" Get peer group based on tickers: 'tkLst'
	passthru of yh_peers()
	help(yh_peers)
	"""
	return yh_peers(tkLst=tkLst,debugTF=debugTF,peersLstTF=peersLstTF)

if __name__ == '__main__':
	from _alan_optparse import parse_opt, subDict
	opts, args = parse_opt(sys.argv)
	df,peerLst,peerInfo = yh_peers(args,debugTF=opts['debugTF'])
	if opts['output'] in ['csv','html','dict']:
		sys.stdout.write("{}".format(df2out(df,output=opts['output'])))
	else:
		sys.stdout.write("{}".format(peerLst))
